package it.polito.elite.ontology;

import it.polito.elite.visualization.ontoviewer.enums.LinkDirectionEnum;
import it.polito.elite.visualization.ontoviewer.enums.NodeTypesEnum;
import it.polito.elite.visualization.ontoviewer.enums.QueryLanguageEnum;
import it.polito.elite.visualization.ontoviewer.models.ConceptData;
import it.polito.elite.visualization.ontoviewer.models.InstanceData;
import it.polito.elite.visualization.ontoviewer.models.RelationData;
import it.polito.elite.visualization.ontoviewer.models.RelationType;
import it.polito.elite.visualization.ontoviewer.userInterfaces.VisualizationOptions;

import java.io.FileOutputStream;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.RDFNode;

import edu.stanford.smi.protege.model.Cls;
import edu.stanford.smi.protege.model.Instance;
import edu.stanford.smi.protege.model.KnowledgeBase;
import edu.stanford.smi.protege.model.SimpleInstance;
import edu.stanford.smi.protegex.owl.ProtegeOWL;
import edu.stanford.smi.protegex.owl.jena.Jena;
import edu.stanford.smi.protegex.owl.jena.JenaOWLModel;
import edu.stanford.smi.protegex.owl.model.NamespaceManager;
import edu.stanford.smi.protegex.owl.model.OWLClass;
import edu.stanford.smi.protegex.owl.model.OWLDatatypeProperty;
import edu.stanford.smi.protegex.owl.model.OWLIndividual;
import edu.stanford.smi.protegex.owl.model.OWLModel;
import edu.stanford.smi.protegex.owl.model.OWLNamedClass;
import edu.stanford.smi.protegex.owl.model.OWLObjectProperty;
import edu.stanford.smi.protegex.owl.model.OWLProperty;
import edu.stanford.smi.protegex.owl.model.OWLRestriction;
import edu.stanford.smi.protegex.owl.model.RDFIndividual;
import edu.stanford.smi.protegex.owl.model.RDFProperty;
import edu.stanford.smi.protegex.owl.model.RDFResource;
import edu.stanford.smi.protegex.owl.model.RDFSClass;
import edu.stanford.smi.protegex.owl.model.RDFSDatatype;
import edu.stanford.smi.protegex.owl.model.RDFSLiteral;
import edu.stanford.smi.protegex.owl.model.RDFSNamedClass;

public class OWLOntologyWrapper extends OntologyAbstraction {

	private OWLModel owlModel;

	private Hashtable<String, RelationType> RelationsCache = new Hashtable<String, RelationType>();

	public OWLOntologyWrapper(KnowledgeBase kb) {
		super();
		// casts the ontology kb to an OWL ontology model
		this.owlModel = (OWLModel) kb;
	}

	public void saveOntology(String pathFname) {
		JenaOWLModel model = (JenaOWLModel) this.owlModel;
		try {
			FileOutputStream f = new FileOutputStream(pathFname);
			Jena.dumpRDF(model.getOntModel(), f);
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	// ///////////////////////////////////// IMPORTED ONTOLOGIS & NAMESPACES

	public boolean isPrefixDefined(String prefix) {
		NamespaceManager ns = this.owlModel.getNamespaceManager();
		return ns.getPrefixes().contains(prefix);
	}

	public void addNameSpace(String prefix, String namespace) {
		NamespaceManager ns = this.owlModel.getNamespaceManager();
		ns.setPrefix(namespace, prefix);
	}

	public boolean isURI_imported(String URI) {
		// checks if the ontology referenced by the uri is
		// in the current import list
		Set<String> set = this.owlModel.getAllImports();
		return set.contains(URI);
	}

	public void addImport(String URI) {
		// imports the ontology into the model
		try {
			// adds the ontology AND the import directive to the owl file
			this.owlModel.getDefaultOWLOntology().addImports(URI); // directive
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	public void confirmLogicViewSupport() {
		// Checks if logic views are supported, if not imports the ontology
			if (!this.isPrefixDefined(VisualizationOptions.s_LogicViewsOntologyPrefix)) {
				this.addNameSpace(VisualizationOptions.s_LogicViewsOntologyPrefix, VisualizationOptions.s_LogicViewsOntology);
			// Importing the data in the current model,  loading the ontology & getting its model
			OWLModel owlModel;
			try {
				owlModel = ProtegeOWL	.createJenaOWLModelFromURI(VisualizationOptions.s_LogicViewsOntology);
				// gets all the AnnotationProperties
				Collection annoPropertiesCollection = owlModel
						.getOWLAnnotationProperties();
				for (Object obj : annoPropertiesCollection) {
					// the current annotation property
					OWLProperty annoProperty = (OWLProperty) obj;
					if (!annoProperty.isIncluded())
						this.addAnnotationProperty(VisualizationOptions.s_LogicViewsOntologyPrefix+ ":" + annoProperty.getName(),
										annoProperty.isObjectProperty());
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	// /////////////////////////// CONCEPTS, RESTRICTIONS & TAXONOMY
	/**
	 * Returns all the classes (concepts) in a knowldge base
	 * 
	 * @param listImportedClasses
	 *            true if also imported classes should be shown
	 * @return a HashSet containing the names of all the classes in the
	 *         knowledge base
	 */
	public HashSet<String> getAllClasses(boolean listImportedClasses) {
		// extracting all classes from the knowledge base
		Collection<RDFSClass> classesCollection = this.owlModel
				.getRDFSClasses();

		// creating the HashSet for containing the classes names
		HashSet<String> classes = new HashSet<String>();

		/*
		 * composing the final HashSet by iterating through the kb classes and
		 * applying filtering rules
		 */
		for (RDFSClass cls : classesCollection) {

			// checks whether it is not a SYSTEM CLASS and is VISIBLE
			// checks if it is imported or not
			if ((!cls.isSystem()) && (cls.isVisible())) {
				if ((listImportedClasses && cls.isIncluded())
						|| (!listImportedClasses && !cls.isIncluded()))
					classes.add(cls.getName());
			}
		}

		// if the resulting HashSet is not empty returns the HashSet otherwise
		// returns an empty set and logs a warning on the console
		if (!classes.isEmpty()) {
			this.log.log(Level.INFO,
					"Successfully retrieved classes, returning a non-empty HashSet");
		} else {
			this.log.log(Level.WARNING,
					"Something bad occurred, perhaps the ontology is empty? returning a empty HashSet");
		}
		return classes;
	}
	
	/**
	 * Retrieves "logical" information about the concept and stores it 
	 * in proper fields of the relative ConceptData: DifferentFrom, SameAs, 
	 * Anonymous Superclasses (unions, intersections)
	 * 
	 * @param clsURI the concept class name
	 * @param data the concept data
	 */
	public void setLogicalConstraints(String clsURI, ConceptData data) {
		try {
			// getting the class
			OWLNamedClass cls=this.owlModel.getOWLNamedClass(clsURI);
			//Getting DISJOINT WITH with and saving them in concept data
			HashSet<String> namedConceptsSet=new HashSet<String>();
			for(Object res: cls.getDisjointClasses()){
				NodeTypeAbstraction node=new NodeTypeOWL((RDFResource) res);
				if(node.getType()==NodeTypesEnum.NamedClass)
					namedConceptsSet.add(node.toString());
			}
			data.setDifferentFrom(namedConceptsSet);
//		Getting anonymous logical parents and saving them in concept data
			for(Iterator it=cls.getSuperclasses(false).iterator();it.hasNext();){
				RDFSClass myCls=(RDFSClass)it.next();
				if(myCls.isAnonymous()){
					NodeTypeAbstraction node=new NodeTypeOWL(myCls);
					if(node.isLogicalClass()){
						data.setAnonymousParent(node);
					}
				}
			}
		} catch (ClassCastException e) {
			System.err.println("!! Warrning: "+clsURI+" is a RDFS class. No OWL features available for it !!");
		}
	}
	
	/**
	 * Retrieves "logical" information about the concept and stores it 
	 * in proper fields of the relative Instance Data
	 * 
	 * @param instanceURI the instance name
	 * @param data the instance data
	 */
	public void setLogicalConstraints(String instanceURI, InstanceData data) {
		try {
			// getting the instance
			OWLIndividual individual =  this.owlModel.getOWLIndividual(instanceURI);
//		Getting DIFFERENT FROM and saving them in instance data
			HashSet<String> namedIndividualsSet=new HashSet<String>();
			for(Object res: individual.getDifferentFrom()){
				NodeTypeAbstraction node=new NodeTypeOWL((RDFResource) res);
				if(node.getType()==NodeTypesEnum.NamedInstance)
					namedIndividualsSet.add(node.toString());
			}
			data.setDifferentFrom(namedIndividualsSet);
			
//		Getting SAME AS and saving them in instance data
			namedIndividualsSet=new HashSet<String>();
			for(Object res: individual.getSameAs()){
				NodeTypeAbstraction node=new NodeTypeOWL((RDFResource) res);
				if(node.getType()==NodeTypesEnum.NamedInstance)
					namedIndividualsSet.add(node.toString());
			}
			data.setSameAs(namedIndividualsSet);
		} catch (ClassCastException e) {
			System.err.println("!! Warrning: "+instanceURI+" is a RDF individual. No OWL features available for it !!");
		}
	}
	
	public Hashtable<String, HashSet<NodeTypeAbstraction>> getRestrictions(
			String clsURI) {
		// Creates an hashset that will contain restrictions, if present
		HashSet<NodeTypeAbstraction> restrictionsSet = new HashSet<NodeTypeAbstraction>();
		// getting the class

		RDFSNamedClass cls = this.owlModel
				.getRDFSNamedClass(clsURI);
		if (cls instanceof OWLNamedClass) {
			Collection<OWLRestriction> restrictions = ((OWLNamedClass) cls)
					.getRestrictions();

			for (OWLRestriction restr : restrictions) {
				NodeTypeAbstraction restriction = new NodeTypeOWL(restr);
				restrictionsSet.add(restriction);
			}
		}

		// grouping restrictions on the same relation
		Hashtable<String, HashSet<NodeTypeAbstraction>> relationGroups = new Hashtable<String, HashSet<NodeTypeAbstraction>>();
		for (NodeTypeAbstraction myRestriction : restrictionsSet) {
			String relURI = myRestriction.getRestrictionPropertyURI();
			if (relationGroups.containsKey(relURI)) {
				HashSet<NodeTypeAbstraction> set = relationGroups.get(relURI);
				set.add(myRestriction);
			} else {
				HashSet<NodeTypeAbstraction> set = new HashSet<NodeTypeAbstraction>();
				set.add(myRestriction);
				relationGroups.put(relURI, set);
			}
		}

		return relationGroups;
	}

	/**
	 * Lists all the root classes of the ontology (leaved here for backward
	 * compatibility)
	 * 
	 * @param listImportedClasses
	 *            true if also imported classes should be shown
	 * @return an Iterator over the names of all the root classes in the
	 *         knowledge base
	 */
	public HashSet<String> getRoots(boolean listImportedClasses) {
		// extracting the root classes from the knowledge base
		RDFSNamedClass thing = this.owlModel.getOWLThingClass();

		// the roots of the ontology are all the direct subclasses of thing
		// that are not SYSTEM classes and that are VISIBLE
		Collection<RDFSNamedClass> rootsCollection = thing.getNamedSubclasses();

		// creating the HashSet for containing the classes names
		HashSet<String> roots = new HashSet<String>();

		// composing the final HashSet by iterating through the kb root classes
		// and applying filtering rules
		for (RDFSNamedClass cls : rootsCollection) {
			// checks whether it is not a SYSTEM CLASS and is VISIBLE
			if ((!cls.isSystem()) && (cls.isVisible())) {
				roots.add(cls.getName());
			}
		}

		// if the resulting HashSet is not empty returns the HashSet otherwise
		// returns an empty set and logs a warning on the console
		if (!roots.isEmpty()) {
			this.log	.log(Level.INFO,	"Successfully retrieved root classes, returning a non-empty HashSet");
		} 
		else {
			this.log
					.log(
							Level.WARNING,
							"Something bad occurred, perhaps the ontology is empty? returning a empty HashSet");
		}
		return roots;
	}

	/**
	 * A method for retrieving all the subclasses of a given class given its
	 * name
	 * 
	 * @param clsName
	 *            the class name
	 * @return an Iterator &lt;String&gt; over the subclasses names
	 */
	public HashSet<String> getSubClasses(String clsName) {
		// getting the class
		RDFSNamedClass cls = this.owlModel.getRDFSNamedClass(clsName);

		// returning the subclasses
		Collection subClassesCollection = cls.getNamedSubclasses();

		// creating a new set item in the required form
		HashSet<String> subClassesNames = new HashSet<String>();

		// filling up the HashSet
		Iterator subClassesIterator = subClassesCollection.iterator();

		while (subClassesIterator.hasNext()) {
			// the subclass
			Cls subCls = (Cls) subClassesIterator.next();

			// if the class is not a system class and is visible
			if ((subCls.isVisible()) && (!subCls.isSystem())) {
				// the subclass name
				String subClsName = subCls.getName();

				// debug
				this.log.log(Level.INFO, " " + subClsName + " subclass of: "
						+ clsName);

				// fills-up the HashSet
				subClassesNames.add(subClsName);
			}

		}
		return subClassesNames;
	}

	/**
	 * A method that provides the number of direct subclasses for a given class
	 * (identified by its name)
	 * 
	 * @param clsName
	 *            the name of the class
	 * @return the number of subclasses as an int
	 */
	public int getNumberOfSubClasses(String clsName) {
		// getting the class
		RDFSNamedClass cls = this.owlModel.getRDFSNamedClass(clsName);
		// returning the number of subclasses
		return cls.getSubclassCount();

	}

	/**
	 * A method for retrieving all the superclasses of a given class given its
	 * name
	 * 
	 * @param clsName
	 *            the class name
	 * @return an Iterator &lt;String&gt; over the superclasses names
	 */
	public HashSet<String> getSuperClasses(String clsName) {
		// getting the class
		RDFSNamedClass cls = this.owlModel.getRDFSNamedClass(clsName);

		// returning the subclasses
		Collection<RDFSNamedClass> superClassesCollection = cls
				.getNamedSuperclasses();

		// creating a new set item in the required form
		HashSet<String> superClassesNames = new HashSet<String>();

		for (RDFSNamedClass superCls : superClassesCollection) {

			// if the class is not a system class and is visible
			if ((superCls.isVisible()) && (!superCls.isSystem())) {
				// the subclass name
				String superClsName = superCls.getName();

				// debug
				this.log.log(Level.INFO, " " + superClsName
						+ " superclass of: " + clsName);

				// fills-up the HashSet
				superClassesNames.add(superClsName);
			}
		}
		return superClassesNames;
	}

	// /////////////////////////// RELATIONS (OBJ & DATATYPE properties)

	/**
	 * Lists all the slots declared in a protege knowledge base, excluding the
	 * imported slots if listImported is true. 
	 * 
	 * @param listImported
	 *            true if slots coming from imported classes shall be shown
	 * @return an Hastable containing the name of the slots together with the
	 *         eventual annotation properties of the slot
	 */

	public Hashtable<String, RelationType> getAllRelations(boolean listImported) {
		// checks if the cache is not already initted
		if (!this.RelationsCache.isEmpty())
			return this.RelationsCache;

		this.RelationsCache = this.getAllRelationsOWL(listImported);
		return this.RelationsCache;
	}
	
	private Hashtable<String, RelationType> getAllRelationsOWL(boolean listImported) {

		// extracts all the ObjectProperties in the owl model
		Collection<OWLProperty> propertiesCollection = this.owlModel.getUserDefinedOWLObjectProperties();
		propertiesCollection.addAll(this.owlModel.getUserDefinedOWLDatatypeProperties());
		
//	  creates the Hashtable for storing results
		Hashtable<String, RelationType> owlProperties = new Hashtable<String, RelationType>();
		
//		 Iterates through all founded ObjectProperties
		for (OWLProperty property : propertiesCollection) {
			if(property.isIncluded() || property.isAnnotationProperty()) 
				continue;
			
			RelationType rType = new RelationType();
			
			rType.setDatatype( !property.isObjectProperty() );
			
			rType.setFunctional( property.isFunctional() );
			
			RDFProperty inverse = property.getInverseProperty();
			if(inverse != null){
				String invPropertyURI = inverse.getLocalName();
				rType.setInverseProperty(invPropertyURI);
			}
			
			if(property.isObjectProperty()){
				OWLObjectProperty objProperty = (OWLObjectProperty)property;
				rType.setTransitive( objProperty.isTransitive() );
				rType.setSymmetric( objProperty.isSymmetric() );
				rType.setInverseFunctional( objProperty.isInverseFunctional() );
			}
			
			owlProperties.put(property.getLocalName(), rType);
		}
		return owlProperties;
	}

	/**
	 * Give the uri of a class, returns an hashtable of relations starting from
	 * the class together with an Hashset of endnodes for each relations
	 * (relationUri, EndNodes)
	 * 
	 * @param uri
	 *            the uri of the starting class
	 * @param direct
	 *            --direct relations only (currently ignored)
	 * @param listImportedClass --
	 *            show system classes or not (to be deprecated)
	 * @return (relationUri &lt;Uri&gt;, EndNodes &lt;HashSet&gt;)
	 */
	public Hashtable<String, HashSet<RelationData>> getObjectProperties(
			String uri, boolean direct, boolean listImportedClass) {
		// creates the Hashtable for returning data
		Hashtable<String, HashSet<RelationData>> objectProperties = new Hashtable<String, HashSet<RelationData>>();
		// casts the class at an RDFS class
		RDFSClass startCls = (RDFSClass) this.owlModel.getRDFSNamedClass(uri);

		// gets all the object properties of kb
		if (this.RelationsCache.isEmpty())
			this.getAllRelations(false);

		// here in the RelationCache there are all the object properties, we
		// should check whether some of them are starting from the current class
		for (String property : this.RelationsCache.keySet()) {
			RelationType rType = this.RelationsCache.get(property);
			if(! rType.isObjectProperty()) continue;
			
			NodeTypeOWL myRange=null;
			boolean foundRange=false;
			RDFResource myRangeRDFS = startCls.getPropertyValueAs(this.owlModel
					.getOWLObjectProperty(property), RDFSClass.class);
			
			if ((myRangeRDFS != null)) {
				// ******************************************************
				// **** OWL FULL Object Properties as Instances ********
				// ******************************************************
				this.log.log(Level.INFO, "identified relation " + property
						+ "value: " + myRangeRDFS.toString());

				myRange = new NodeTypeOWL(myRangeRDFS);
				foundRange=true;
			}
			else {
				// ************************************
				// ****** Try OWL DL properties *******
				// ************************************

				// getting the OWLDLObjectProperty
				OWLObjectProperty owlDLProperty = this.owlModel
						.getOWLObjectProperty(property);

				// getting the property domain (the boolean parameter sets if
				// superproperties must be included or not)
				Collection domainClses = owlDLProperty.getDomains(false);

				// checking if the current class is in the property domain
				if (domainClses.contains(startCls)) {
					// extracts the range of the property
					RDFResource rangeRes = owlDLProperty.getRange();

					if (rangeRes instanceof RDFSClass) {
						myRange = new NodeTypeOWL(rangeRes);
						foundRange=true;
					}
				}
			}
			//Storing the info
			if(myRange!=null && foundRange){
				if (objectProperties.containsKey(property)) {
					// if the relation is already in the HashTable add the current node to the range set associated to that obj prop
					HashSet<RelationData> relationDataSet = objectProperties.get(property);
					RelationData rdata = new RelationData(property,	LinkDirectionEnum.out, rType);
					rdata.setRelatedNode(myRange);
					// adding the current node
					relationDataSet.add(rdata);
					
				}

				else {
					// if the relation is not already in the HashTable creates a new entry in the Hashtable and then 
					//adds the current nodeType to the set associated to the relation
					HashSet<RelationData> relationDataSet = new HashSet<RelationData>(1, 2);
					// adding the current relation data
					RelationData rdata = new RelationData(property,	LinkDirectionEnum.out, rType );
					rdata.setRelatedNode(myRange);
					relationDataSet.add(rdata);
					// storing the HashSet in the Hashtable
					objectProperties.put(property, relationDataSet);
				}
				
				this.log.log(Level.INFO, 	"The relation is already existing!: ONLY ADDED "
								+ myRange.toString());
			}

		}

		return objectProperties;
	}

	/**
	 * A method that, given the name of a concept, provides as result a Hastable
	 * containing the couples datatype property, xsd type for each datatype
	 * property which has the given concept as domain.
	 * 
	 * @param concept
	 * @return the hashtable of datatype properties for concept together with
	 *         their xsd type
	 */
	public Hashtable<String, HashSet<RelationData>> getDatatypeProperties(	String concept) {
		// preparing the set of properties
		Hashtable<String, HashSet<RelationData>> datatypeProperties = 
			new Hashtable<String, HashSet<RelationData>>();

		// getting the concept class
		RDFSClass owlCls = this.owlModel.getRDFSNamedClass(concept);

		// searching all the datatype properties having owlCls in their domain
		Collection<OWLDatatypeProperty> datatypePropertiesCollection = this.owlModel
				.getUserDefinedOWLDatatypeProperties();

		// iterating over the collection of datatype properties
		for (OWLDatatypeProperty datatypeProperty : datatypePropertiesCollection) {

			// cheks if the current class is in the property domain
			Collection domainClses = datatypeProperty.getDomains(false);
			if (domainClses.contains(owlCls)) {
				// the property name
				String datatypePropertyName = datatypeProperty.getName();

				// the property range value
				RDFSDatatype datatypePropertyRange = datatypeProperty
						.getRangeDatatype();

				// if not null gets the name
				if (datatypePropertyRange != null) {
					// adds the property to the set ONLY 1 DATATYPE RANGE is possible, SO...
					HashSet<RelationData> relDataSet = new HashSet<RelationData>(1);
					RelationData rdata = new RelationData(datatypePropertyName,	LinkDirectionEnum.out, this.RelationsCache.get(datatypePropertyName) );
					rdata.setRelatedNode(new NodeTypeOWL(datatypePropertyRange));
					relDataSet.add(rdata);
					datatypeProperties.put(datatypePropertyName, relDataSet);

					// debug
					this.log.log(Level.WARNING, "found datatype property for "
							+ concept + ": " + datatypePropertyName + "["
							+ datatypePropertyRange + "]");
				}
			}
		}

		return datatypeProperties;
	}

	// /////////////////////////////////// ANNOTATIONS

	/**
	 * A method that returns an hashtable containing all the annotations
	 * relative to the input class
	 * 
	 * @param resourceURI
	 *            the class or propery we want to extract annotation properties from
	 * @return an Hashtable containing property name and value
	 */
	public Hashtable<String, Object> getAnnotationProperties(String resourceURI, boolean isNamedClass) {

		// Creates the hastable that will contain annotations, if present
		Hashtable<String, Object> result = new Hashtable<String, Object>();

		// getting the class or the property
		OWLProperty prop = null;
		RDFSNamedClass cls = null;
		if(isNamedClass)
			cls = this.owlModel.getRDFSNamedClass(resourceURI);
		else
			prop = this.owlModel.getOWLProperty(resourceURI);
		
		
		// gets all the AnnotationProperties
		Collection<OWLProperty> annoPropertiesCollection = this.owlModel.getOWLAnnotationProperties();

		// for each annotation property extract the value for the given Object Property
		for (OWLProperty annoProperty : annoPropertiesCollection) {

			// checks if the objectPropertie being examined has a value for this annotation property
			Object annoPropertyValue = null;
			if(isNamedClass && cls != null)
				annoPropertyValue = cls.getPropertyValue(annoProperty);
			else if(!isNamedClass && prop != null){
				annoPropertyValue = prop.getPropertyValue(annoProperty);
			}
			else continue;
			
			// if the value is not null can insert the entry into the hastable of annotation properties
			if (annoPropertyValue != null) {
				// adding the annotation property to the Hashtable
				result.put(annoProperty.getName(), annoPropertyValue);
			}
		}
		return result;
	}

	/**
	 * Remove an annotation property of type <i>property<i>, having <i>object</i>
	 * as value, referred to a given concept (OWL class)
	 * 
	 * @param clsURI
	 *            the URI of the concept or the object property with the annotation
	 * @param property	       The name of the annotation property we want to remove
	 * @param value		        The value of the annotation
	 * @return true on success, false if the removal fails
	 */
	public boolean removeAnnotationProperty4URI(String clsURI, String property,
			Object value) {
		boolean result = false;

		// casts the class at an OWL class
		OWLClass cls = (OWLClass) this.owlModel.getOWLNamedClass(clsURI);
		OWLObjectProperty objProp = this.owlModel.getOWLObjectProperty(clsURI);
		
		// search the annotation property and remove it
		for (Object obj : this.owlModel.getOWLAnnotationProperties()) {
			OWLProperty prop = (OWLProperty) obj;
			if (prop.getName().equals(property)) {
				if(cls != null){
					cls.removePropertyValue(prop, value);
					result = true;
				}
				else if(objProp != null){
					objProp.removePropertyValue(prop, value);
					result = true;
				}
				else
					result = false;
			}
		}
		return result;
	}

	/**
	 * Adds an annotation property of type <i>property<i>, having <i>object</i>
	 * as value, referred to a given concept (OWL class)
	 * 
	 * @param uri
	 *            the URI of the concept
	 * @param property
	 *            The name of the annotation property we want to add
	 * @param value
	 *            The value of the annotation
	 * @return true on success, false if the operation fails
	 */
	public boolean addAnnotationProperty4URI(String uri, String property,
			Object value) {
		boolean result = false;
		// get the corresponding resource
		RDFResource resource = this.owlModel.getRDFResource(uri);

		// search the annotation property and add the value
		for (Object obj : this.owlModel.getOWLAnnotationProperties()) {
			OWLProperty prop = (OWLProperty) obj;
			if (prop.getName().equals(property)) {
				if (resource instanceof OWLClass) {
					OWLClass cls = (OWLClass) resource;
					cls.addPropertyValue(prop, value);
					result = true;
				} else if (resource instanceof RDFProperty) {
					RDFProperty myProp = (RDFProperty) resource;
					myProp.addPropertyValue(prop, value);
					result = true;
				}

				break;
			}
		}
		return result;
	}

	public void addAnnotationProperty(String property, boolean isObjectProperty) {

		if (isObjectProperty)
			this.owlModel.createAnnotationOWLObjectProperty(property);
		else
			this.owlModel.createAnnotationOWLDatatypeProperty(property);
	}

	// /////////////////////////////////// INDIVIDUALS
	/**
	 * A method for getting all the instances of a concept in the ontology (for
	 * a OWL model)
	 * 
	 * @return the HashSet containing all the instances of a given ontology
	 */
	public HashSet<String> getAllInstances() {
		// the HashSet of all instances
		HashSet<String> instances = new HashSet<String>();

		// getting all instances from the model
		Collection individualsCollection = this.owlModel.getRDFIndividuals();

		// getting the iterator over the instances collection
		Iterator individualsIterator = individualsCollection.iterator();

		// iterating through the available instances for extracting their names
		while (individualsIterator.hasNext()) {
			// the owl instance object
			// OWLIndividual individual = (OWLIndividual)
			// individualsIterator.next();
			RDFIndividual individual = (RDFIndividual) individualsIterator
					.next();

			// checking that the instance is visible and is not a system
			// instance
			if ((individual.isVisible()) && (!individual.isSystem())) {
				// the instance name
				String individualName = individual.getName();

				// storing the instances
				instances.add(individualName);

				// debug
				this.log.log(Level.INFO, "Found instance: " + individualName);
			}
		}

		return instances;
	}

	public HashSet<String> getInstancesOfConcept(String concept) {
		// create a set 4 results
		HashSet<String> result = new HashSet<String>();

		// gets the class object
		Cls cls = this.owlModel.getCls(concept);

		// checks if there are instances for that class
		Collection instancesCollection = cls.getDirectInstances();

		// gets an iterator over the instances
		Iterator instancesIterator = instancesCollection.iterator();

		// search the first simple instance
		while (instancesIterator.hasNext()) {
			// gets the current instance
			Instance instance = (Instance) instancesIterator.next();

			// checks if the istance is a simple instance
			if ((instance instanceof SimpleInstance) && (instance.isVisible())
					&& (!instance.isSystem())) {
				result.add(instance.getName());
			}
		}
		return result;
	}

	/**
	 * A method for checking if a given concept c has some instances or not
	 * 
	 * @param concept
	 *            the concept for which instances might exist
	 * @return true if it has some instances, false otherwise
	 */
	public boolean hasInstances(String concept) {
		// gets the class object
		RDFSNamedClass cls = this.owlModel.getRDFSNamedClass(concept);

		// checks if there are instances for that class
		Collection instancesCollection = cls.getInstances(true);

		// gets an iterator over the instances
		Iterator instancesIterator = instancesCollection.iterator();

		// search the first simple instance
		while (instancesIterator.hasNext()) {
			// gets the current instance
			Instance instance = (Instance) instancesIterator.next();

			// checks if the istance is a simple instance
			if ((instance instanceof SimpleInstance) && (instance.isVisible())
					&& (!instance.isSystem())) {
				// is the instance is a simple istance true shall be returned
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns an Hashtable containing for each instance "linked" to the
	 * instance i, a Set of objectProperties that start from i and reach the
	 * "linked" instance (usually such set is composed by only one
	 * objectProperty).
	 * 
	 * @param instance     the instance i
	 * @return the hashtable of linked instances
	 */
	public Hashtable<String, HashSet<NodeTypeAbstraction>> getInstancesObjectProperties(
			String instance) {
		// preparing the object properties hashtable for instances
		Hashtable<String, HashSet<NodeTypeAbstraction>> instancesObjectProperties = new Hashtable<String, HashSet<NodeTypeAbstraction>>();

		// gets the class corresponding to the concept
		RDFIndividual owlInd = this.owlModel.getRDFIndividual(instance);

		// gets all the objetc properties
		if (this.RelationsCache.isEmpty())
			this.getAllRelations(false);

		// here in the RelationCache there are all the object properties, we
		// should check whether some of them are starting from the current class
		for (String property : this.RelationsCache.keySet()) {
			RelationType rType = this.RelationsCache.get(property);
			if(rType.isDatatype())
				continue;
			// getting the property object
			OWLProperty objectProperty = this.owlModel.getOWLProperty(property);

			// gets the value of the property for this class
			Collection<RDFIndividual> objectPropertyValues = owlInd
					.getPropertyValuesAs(objectProperty, RDFIndividual.class);

			// checks wheter the collection is empty or null
			if ((objectPropertyValues != null)
					&& (!objectPropertyValues.isEmpty())) {

				// iterates through the collection of the property values
				for (RDFIndividual endIndividual : objectPropertyValues) {
					// getting the individual name
					NodeTypeOWL myRange = new NodeTypeOWL(endIndividual);

					// storing the data
					if (instancesObjectProperties.containsKey(property)) {
						/*
						 * if the relation is already in the HashTable add the
						 * current node to that relation set
						 */
						HashSet<NodeTypeAbstraction> relationsSet = instancesObjectProperties
								.get(property);

						// addin the current relation
						relationsSet.add(myRange);

						// debug
						this.log.log(Level.INFO,"The relation is already existing!: ONLY ADDED "
										+ myRange.toString());
					} else {
						/*
						 * if the relation endpoint is not already in the
						 * HashTable creates a new entry in the endClass
						 * Hashtable and then adds the current property to the
						 * relation set associated to the newly inserted class
						 */
						HashSet<NodeTypeAbstraction> relationsSet = new HashSet<NodeTypeAbstraction>(
								2, 1);

						// adding the current relation
						relationsSet.add(myRange);

						// storing the HashSet in the Hashtable
						instancesObjectProperties.put(property, relationsSet);

						// debug
						this.log.log(Level.INFO,	"The rel is NOT existing!: ADDED A NEW ENTRY "
										+ property);
					}

				}
			}
		}
		return instancesObjectProperties;
	}


	public Hashtable<String, HashSet<NodeTypeAbstraction>> getInstancesDatatypeProperties(
			String instance) {
		// prepares the hashtable of datatype properties
		Hashtable<String, HashSet<NodeTypeAbstraction>> instancesDatatypeProperties = new Hashtable<String, HashSet<NodeTypeAbstraction>>();

		// gets the individual corresponding to "instance"
		// OWLIndividual individual = ontologyOWL.getOWLIndividual(instance);
		RDFIndividual individual = this.owlModel.getRDFIndividual(instance);

		// searching all the datatype properties starting from the individual in
		// their domain
		Collection<OWLDatatypeProperty> datatypePropertiesCollection = this.owlModel
				.getUserDefinedOWLDatatypeProperties();

		// iterating over the collection of datatype properties
		for (OWLDatatypeProperty datatypeProperty : datatypePropertiesCollection) {

			// checks whether the current instance has or not the current
			// property, and with which value
			Collection<RDFSLiteral> objectDatatypePropertyValuesCollection = individual
					.getPropertyValueLiterals(datatypeProperty);

			// if the property actually exist
			if (objectDatatypePropertyValuesCollection != null
					&& objectDatatypePropertyValuesCollection.size() > 0) {

				// gets the property values as strings
				HashSet<NodeTypeAbstraction> propertyValues = new HashSet<NodeTypeAbstraction>(
						objectDatatypePropertyValuesCollection.size());

				for (RDFSLiteral propertyValue : objectDatatypePropertyValuesCollection)
					propertyValues.add(new NodeTypeOWL(propertyValue));

				// gets the property name
				String datatypePropertyName = datatypeProperty.getName();

				// stores the datatype property name and value
				instancesDatatypeProperties.put(datatypePropertyName,
						propertyValues);

				// debug
				this.log.log(Level.INFO, "Found DatatypeProperty: "
						+ datatypePropertyName + "[" + propertyValues + "]");
			}
		}
		return instancesDatatypeProperties;
	}

	@Override
	public NodeTypeAbstraction createNodeTypeAbstraction(NodeTypesEnum t,
			String expression) {
		return NodeTypeOWL.parseExpression(t, expression);
	}
	public boolean canQuery()
	{
		return true;
	}
	
	public HashSet <NodeTypeAbstraction> query(String queryString, QueryLanguageEnum lang, int queryType)
	{
		HashSet <NodeTypeAbstraction> results = null;
		
		//first check if the query language is uderstood
		if(lang == QueryLanguageEnum.SPARQL)
		{
			//the result HashSet
			results = new HashSet <NodeTypeAbstraction> (); 
			
			//gets a Jena model
			Model jenaModel = this.owlModel.getJenaModel();
			
			//automatically includes all the available prefixes
			StringBuffer prefixesDeclaration = new StringBuffer();
			Map prefixes = jenaModel.getNsPrefixMap();
			Set prefixesSet = prefixes.keySet();
			for(Iterator prefixIterator = prefixesSet.iterator(); prefixIterator.hasNext();)
			{
				//the current prefix
				String prefix = (String) prefixIterator.next();
				
				//debug
				this.log.log(Level.INFO, "prefix "+prefix+":<"+prefixes.get(prefix)+">");
				
				//adding prefixes
				prefixesDeclaration.append("prefix "+prefix+":<"+prefixes.get(prefix)+">");
			}
			
			//adds the prefixes
			queryString = prefixesDeclaration.toString()+" "+queryString;
			
			//prepares the query
			Query query = QueryFactory.create(queryString);
			
			//creates the query execution environment
			QueryExecution qExec = QueryExecutionFactory.create(query,jenaModel);
			
			try
			{
				Object res;
				
				//manage the query type
				switch (queryType)
				{		
					case Query.QueryTypeConstruct:
					{
						res = qExec.execConstruct();
						break;
					}
					case Query.QueryTypeDescribe:
					{
						res = qExec.execDescribe();
						break;
					}
					case Query.QueryTypeSelect:
					{
						System.out.println("Executing query...");
						
						res = qExec.execSelect();
						
						System.out.println("ResultSetSize: "+((ResultSet)res).getRowNumber());
						
						while(((ResultSet)res).hasNext())
						{
							QuerySolution qSolution = ((ResultSet)res).nextSolution();
							Iterator varIterator = qSolution.varNames();
							while(varIterator.hasNext())
							{
								String variable = (String)varIterator.next();
								RDFNode varNode = qSolution.get(variable);
								//NodeTypeAbstraction absNode = new NodeTypeOWL((RDFResource)varNode);
								//results.add(absNode);
								String relativeRes[] = varNode.toString().split("#");
								RDFResource rdfRes = this.owlModel.getRDFResource(relativeRes[1]);
								NodeTypeAbstraction absNode = new NodeTypeOWL(rdfRes);
								results.add(absNode);
								System.out.println("Found RDFRes: "+rdfRes);
								System.out.println("Found RDFNode: "+varNode.toString());
							}
						}
						break;
					}
					
				}
				
			}
			catch(Exception e)
			{
				System.out.println("Error while executing SPARQL query: "+e);
			}
			
			
			
		}
		return results;
	}

	@Override
	public NodeTypeAbstraction getPropertyRange(String propertyURI)  { 
		RDFResource range;
		// getting the OWLDLObjectProperty
			try {
				OWLObjectProperty owlDLProperty = this.owlModel.getOWLObjectProperty(propertyURI);
				range = owlDLProperty.getRange(false);
			} catch (ClassCastException e) {
				OWLDatatypeProperty owlDLProperty = this.owlModel.getOWLDatatypeProperty(propertyURI);
				range = owlDLProperty.getRange(false);
			}
			return new NodeTypeOWL(range);
	}

	
	public OWLModel getModel()
	{
		return this.owlModel;
	}

	@Override
	public RelationType getRelationType(String relURI) {
		return this.RelationsCache.get(relURI);
	}
}
