package it.polito.elite.ontology;

import it.polito.elite.visualization.ontoviewer.enums.NodeTypesEnum;
import it.polito.elite.visualization.ontoviewer.models.ConceptData;
import it.polito.elite.visualization.ontoviewer.models.InstanceData;
import it.polito.elite.visualization.ontoviewer.models.RelationData;
import it.polito.elite.visualization.ontoviewer.models.RelationType;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.smi.protege.model.KnowledgeBase;
import edu.stanford.smi.protegex.owl.model.OWLModel;

public abstract class OntologyAbstraction {
	
	protected Logger log;
	
	protected OntologyAbstraction(){
		this.log = Logger.getLogger("it.polito.elite.ontology");
		// setting the defaul level at INFO
		this.log.setLevel(Level.OFF);
	}
	
	/**
	 * It is sufficient that it handles NamedClasses and NamedInstances
	 * @param t type
	 * @param expression the expression (at now the URI of concept/individual)
	 * @return NodeTypeAbstraction
	 */
	abstract public NodeTypeAbstraction createNodeTypeAbstraction(NodeTypesEnum t, String expression);
	
	static public OntologyAbstraction createOntologyAbstraction(KnowledgeBase kb) {

		OntologyAbstraction onto = null;
		
		// checking if this is an OWL model
		if (kb instanceof OWLModel) {
			//If the protege model is an owl model logs the information and
			// sets the field owl at true (used in owl-related methods)
			onto=new OWLOntologyWrapper(kb);
			onto.log.log(Level.WARNING, "This is an OWL model");
		} else {
			// If the protege model is not an owl model logs the information and
			// sets the field owl at false (used in owl-related methods)
			//onto=new FrameLogicOntologyWrapper(kb);
			System.err.println("Only OWL supported at the moment. This is NOT an OWL model");
			System.exit(-1);
		}
		
		return onto;
	}
	
	/**
	 * Returns all the classes (concepts) in a knowldge base
	 * 
	 * @param listImportedClasses        true if also imported classes should be shown
	 * @return a HashSet containing the names of all the classes in the knowledge base
	 */
	abstract public HashSet<String> getAllClasses(boolean listImportedClasses);
	
	/**
	 * Lists all the root classes of the ontology (leaved here for backward
	 * compatibility)
	 * 
	 * @param listImportedClasses    true if also imported classes should be shown
	 * @return an Iterator over the names of all the root classes in the knowledge base
	 */
	abstract public HashSet<String> getRoots(boolean listImportedClasses);
	
	/**
	 * Lists all the slots declared in a protege knowledge base, excluding the
	 * imported slots if listImported is true. (In this version, listImported is
	 * actually ignored)
	 * 
	 * @param listImported  true if slots coming from imported classes shall be shown
	 * @return an Hastable containing the name of the slots together with information about the
	 * relation type (as if the relation if functional or transitive or symmetric...)
	 */
	abstract public Hashtable<String, RelationType> getAllRelations(boolean listImported);
	
	/**
	 * A method for retrieving all the subclasses of a given class given its name
	 * 
	 * @param clsName  the class name
	 * @return an Iterator &lt;String&gt; over the subclasses names
	 */
	abstract public HashSet<String> getSubClasses(String clsName);
	
	/**
	 * A method that provides the number of direct subclasses for a given class
	 * (identified by its name)
	 * 
	 * @param clsName      the name of the class
	 * @return the number of subclasses as an int
	 */
	abstract public int getNumberOfSubClasses(String clsName); 
	
	/**
	 * A method for retrieving all the superclasses of a given class given its name
	 * 
	 * @param clsName     the class name
	 * @return an Iterator &lt;String&gt; over the superclasses names
	 */
	abstract public HashSet<String> getSuperClasses(String clsName);
	
	/**
	 * A method for getting all the instances  in the ontology (for a RDF/RDFS model)
	 * 
	 * @return the HashSet containing all the instances of a given ontology
	 */
	abstract public HashSet<String> getAllInstances();
	
	/**
	 * A method for getting all the instances of a concept in the ontology (for
	 * a RDF/RDFS model)
	 *
	 * @param concept The concept URI
	 * @return the HashSet containing all the instances of a given concept
	 */
	abstract public HashSet<String> getInstancesOfConcept(String concept);
	
	/**
	 * A method for checking if a given concept c has some instances or not
	 * 
	 * @param concept  the concept for which instances might exist
	 * @return true if it has some instances, false otherwise
	 */
	abstract public boolean hasInstances(String concept);
	
	/**
	 * A method that tells if a relation is towards a literal 
	 * @return the RelationType
	 */
	abstract public RelationType getRelationType(String relURI);
	/**
	 * A method that returns an hashtable containing all the annotations
	 * relative to the input class
	 * 
	 * @param resourceURI   the class or the property we want to extract annotation properties from
	 * @return an Hashtable containing property name and value
	 */
	abstract public Hashtable<String, Object> getAnnotationProperties(String resourceURI, boolean isNamedClass);
	
	/**
	 * Adds an annotation property of type <i>property<i>, having <i>object</i>
	 * as value, referred to a given concept (OWL class)
	 * 
	 * @param uri
	 *            the URI of the concept
	 * @param property
	 *            The name of the annotation property we want to add
	 * @param value
	 *            The value of the annotation
	 * @return true on success, false if the operation fails
	 */
	abstract public boolean addAnnotationProperty4URI(String uri, String property, Object value);
	
	/**
	 * Remove an annotation property of type <i>property<i>, having <i>object</i>
	 * as value, referred to a given concept (OWL class)
	 * 
	 * @param clsURI
	 *            the URI of the concept
	 * @param property
	 *            The name of the annotation property we want to remove
	 * @param value
	 *            The value of the annotation
	 * @return true on success, false if the removal fails
	 */
	abstract public boolean removeAnnotationProperty4URI(String clsURI, String property,  Object value);
	 
	/**
	  * Retrieves restrictions specified for the concept identified by a given URI
	  * @param clsURI The URI of the concept
	  * @return an HashSet of NodeTypeAbstraction Objects
	  */
	abstract public Hashtable<String, HashSet<NodeTypeAbstraction>> getRestrictions(String clsURI);
	
	abstract public NodeTypeAbstraction getPropertyRange(String uri) ;
	
	/**
	 * Give the uri of a class, returns an hashtable of relations starting from
	 * the class together with an Hashset of endnodes for each relations
	 * (relationUri, EndNodes)
	 * 
	 * @param uri     the uri of the starting class
	 * @param direct   --direct relations only (currently ignored)
	 * @param listImportedClass --  show system classes or not (to be deprecated)
	 * @return (relationUri &lt;Uri&gt;, EndNodes &lt;HashSet&gt;)
	 */
	abstract public Hashtable<String, HashSet<RelationData>> getObjectProperties(String uri,
			boolean direct, boolean listImportedClass);
	
	/**
	 * A method that, given the name of a concept, provides as result a Hastable
	 * containing the couples datatype property, xsd type for each datatype
	 * property which has the given concept as domain.
	 * 
	 * @param concept
	 * @return the hashtable of datatype properties for concept together with
	 *         their xsd type
	 */
	abstract public Hashtable<String, HashSet<RelationData>> getDatatypeProperties(String concept);
	
	/**
	 * Returns an Hashtable containing for each instance "linked" to the
	 * instance i, a Set of objectProperties that start from i and reach the
	 * "linked" instance (usually such set is composed by only one
	 * objectProperty).
	 * 
	 * @param instance
	 *            the instance i
	 * @return the hashtable of linked instances
	 */
	abstract public Hashtable<String, HashSet<NodeTypeAbstraction>> getInstancesObjectProperties(String instance);
	

	abstract public Hashtable<String, HashSet<NodeTypeAbstraction>> getInstancesDatatypeProperties(String instance);
	
	abstract public void confirmLogicViewSupport();
	
	abstract public void setLogicalConstraints(String clsURI, ConceptData data);
	
	abstract public void setLogicalConstraints(String instanceURI, InstanceData data);
	/**
	 * Sets the logging level for the OntologyAbstraction class, allowed levels
	 * include:
	 * <ul> <li>INFO (default)</li><li>OFF</li><li>FINE</li>
	 * <li>FINEST</li><li>WARNING</li><li>SEVERE</li> </ul>
	 * 
	 * @param lv - The logging level
	 */
	public void setLoggingLevel(String lv) {
		this.log.setLevel(Level.parse(lv));
	}
	
	abstract public boolean canQuery();
}
