package it.polito.elite.visualization.ontoviewer.models;

import it.polito.elite.ontology.NodeTypeAbstraction;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;
import java.util.Map.Entry;

import javax.swing.tree.DefaultMutableTreeNode;

/**
 * 
 * ConceptData class summarizes useful information (structural, semantic and positional)
 * related to a given concept. 
 * @author Alessio
 *
 */
public class ConceptData{
	
	/**
	 * 	The table contains semantic arcs toward other concepts 	
	 * each table entry consists of a String key as the relation uri  
	 * and a set of RelationData as Object
	 */
	protected Hashtable<String,HashSet<RelationData>> outgoingLinks; 
	protected Hashtable<String,HashSet<RelationData>> incomingLinks;
	protected HashSet<String> differentFrom;
	protected HashSet<RelationData> selfLoopLinks;
	protected Hashtable<String,HashSet<RelationData>> datatypeLinks;
	protected NodeTypeAbstraction anonymousParent=null;//logical expression as UNION or INTERSECTION..
	protected String uri = null;
	
	
	private float conceptSphereRadius=0;
	private String parentURI=null;;
	private HashSet<String> adoptiveParentsURI=null;
	private HashSet<String> instancesURI=null;
	private Hashtable<String,Integer> childrenURI=null;
	private boolean hasNonDirectInstances=false;
	
	/**
	 * base element for constructing the taxonomy tree
	 */
	private DefaultMutableTreeNode node;
	
	public ConceptData(String uri){
		incomingLinks = new Hashtable<String,HashSet<RelationData>>();
		outgoingLinks = new Hashtable<String,HashSet<RelationData>>();
		selfLoopLinks = new HashSet<RelationData>();
		differentFrom = new HashSet<String>();
		datatypeLinks = new Hashtable<String,HashSet<RelationData>>();
		this.uri = uri;
		childrenURI=new Hashtable<String,Integer>();
		adoptiveParentsURI=new HashSet<String>(2);
	}
	
	public String getParent() {
		return this.parentURI;
	}
	
	public void setParent(String s) {
		this.parentURI = s;
	}
	
	public void addChild(String uri, int numChildren) {
		this.childrenURI.put(uri,new Integer(numChildren));
	}
		
	public void addAdoptiveParent(String uri) {
		this.adoptiveParentsURI.add(uri);
	}
	
	public Set<String> getAdoptiveParents() {
		return adoptiveParentsURI;
	}
	
	public HashSet<String> getAllParents(){
		HashSet<String> parents=new HashSet<String>(this.adoptiveParentsURI);
		if(this.parentURI!=null)
			parents.add(this.parentURI);
		return parents;
	}
	
	public int getNumberOfParents(){
		int k=0;
		if(this.parentURI!=null)
			k=1;
		return this.adoptiveParentsURI.size()+k;
	}
	
	public Set<String> getChildren(){
		return childrenURI.keySet();
	}
	
	/**
	 * Returns an array of Uris ordered by the respective number of their Children
	 */	
	public OrderedURIsElm[] getOrderedChildrenURIs()
	{
		OrderedURIsElm[] array=new OrderedURIsElm[childrenURI.size()];
		Set<Entry<String,Integer>> set=childrenURI.entrySet();
		int i=0;
		for(Entry<String,Integer> entry:set){
			array[i++]=new OrderedURIsElm(entry.getKey(),entry.getValue());
		}
		Arrays.sort(array);
		return array;
	}
	
			
	public int getLayer() {
		return this.node.getLevel();
	}

	/**
	 * @return Returns the childNumber.
	 */
	public int getNumberOfChildren() {
		return this.childrenURI.size();
	}

	public int getSubTreeDepth() {
		return this.node.getDepth();
	}

	/**
	 * @return Returns the conceptSphereRadius.
	 */
	public float getConceptSphereRadius() {
		return conceptSphereRadius;
	}

	/**
	 * @param conceptSphereRadius The conceptSphereRadius to set.
	 */
	public void setConceptSphereRadius(float conceptSphereRadius) {
		this.conceptSphereRadius = conceptSphereRadius;
	}

	/**
	 * @return Returns the hasInstances.
	 */
	public boolean hasInstances() {
		return (instancesURI.size()>0)?true:false;
	}

	/**
	 * @return Returns the instancesURI.
	 */
	public Set<String> getInstancesURI() {
		return instancesURI;
	}

	/**
	 * @param instancesURI The instancesURI to set.
	 */
	public void setInstancesURI(HashSet<String> instancesURI) {
		this.instancesURI = instancesURI;
	}

	/**
	 * @return Returns the node.
	 */
	public DefaultMutableTreeNode getNode() {
		return node;
	}

	/**
	 * @param node The node to set.
	 */
	public void setNode(DefaultMutableTreeNode node) {
		this.node = node;
	}

	public boolean hasNonDirectInstances() {
		return hasNonDirectInstances;
	}

	public void setNonDirectInstances(boolean hasNonDirectInstances) {
		this.hasNonDirectInstances = hasNonDirectInstances;
	}

	public void addOutgoingLinks(Hashtable<String,HashSet<RelationData>> t) {
		this.outgoingLinks.putAll(t);	
	}
	
	public void removeOutgoingLink(String destUri){
		outgoingLinks.remove(destUri);
	}
	
	public void addDatatypeLinks(Hashtable<String,HashSet<RelationData>> t) {
		this.datatypeLinks.putAll(t);	
	}
		
	public void addIncomingLink(String relURI, RelationData relData){
		if(this.incomingLinks.containsKey(relURI)){
			HashSet<RelationData> set=this.incomingLinks.get(relURI);
			set.add(relData);
		}
		else{
			HashSet<RelationData> set=new HashSet<RelationData>(1);
			set.add(relData);
			this.incomingLinks.put(relURI, set);
		}
	}
	
	/**
	 * @return Returns the datatypeLinks.
	 */
	public Hashtable<String, HashSet<RelationData>> getDatatypeLinks() {
		return datatypeLinks;
	}

	/**
	 * @param datatypeLinks The datatypeLinks to set.
	 */
	public void setDatatypeLinks(Hashtable<String, HashSet<RelationData>> datatypeLinks) {
		this.datatypeLinks = datatypeLinks;
	}

	/**
	 * @return Returns the incomingLinks.
	 */
	public Hashtable<String, HashSet<RelationData>> getIncomingLinks() {
		return incomingLinks;
	}

	/**
	 * @param incomingLinks The incomingLinks to set.
	 */
	public void setIncomingLinks(Hashtable<String, HashSet<RelationData>> incomingLinks) {
		this.incomingLinks = incomingLinks;
	}

	/**
	 * @return Returns the outgoingLinks.
	 */
	public Hashtable<String, HashSet<RelationData>> getOutgoingLinks() {
		return outgoingLinks;
	}

	/**
	 * @param outgoingLinks The outgoingLinks to set.
	 */
	public void setOutgoingLinks(Hashtable<String, HashSet<RelationData>> outgoingLinks) {
		this.outgoingLinks = outgoingLinks;
	}

	/**
	 * @return Returns the uri.
	 */
	public String getUri() {
		return uri;
	}

	/**
	 * @return Returns the selfLoopLinks.
	 */
	public HashSet<RelationData> getSelfLoopLinks() {
		return selfLoopLinks;
	}


	public void addSelfLoopLink(RelationData relation) {
		this.selfLoopLinks.add(relation);
	}

	public NodeTypeAbstraction getAnonymousParent() {
		return anonymousParent;
	}

	public void setAnonymousParent(NodeTypeAbstraction anonymousParent) {
		this.anonymousParent = anonymousParent;
	}
	
	public boolean hasAnonymousParent(){
		if (this.anonymousParent==null)
			return false;
		else
			return true;
	}
	
	public HashSet<String> getDifferentFrom() {
		return differentFrom;
	}

	public void setDifferentFrom(HashSet<String> differentFrom) {
		this.differentFrom = differentFrom;
	}

}