package it.polito.elite.visualization.ontoviewer.tools;

import it.polito.elite.visualization.ontoviewer.userInterfaces.VisualizationOptions;

import java.awt.Font;

import javax.media.j3d.Appearance;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.Font3D;
import javax.media.j3d.FontExtrusion;
import javax.media.j3d.Material;
import javax.media.j3d.OrientedShape3D;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Text3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.TransparencyAttributes;
import javax.vecmath.Color3f;
import javax.vecmath.Point3f;

import com.sun.j3d.utils.geometry.Sphere;

/**
 * This class offers an utility to draw spheres as a building block for scenes.
 * It exposes a dummy constructor, a group of set methods in order to explicitly
 * set characteristics as transparency, the radius.. It provides also a method
 * createColoredSphere which return a java 3D TransformGroup that will be
 * included by the scene itself. An example of its use is: TransformGroup tgx =
 * new TransformGroup(); String label="..."; ConceptSphere sfera = new
 * ConceptSphere(); sfera.setRadius(radius); Color3f color=new
 * Color3f(Color.WHITE); tgx.addChild(sfera.createColoredSphere(....));
 * 
 * @author alessio.bosca@polito.it
 * 
 */
public class ConceptSphere {

	// used as the UserData associated to the 3d obj on the screen
	private String sphereLabel = null;

	private float labelScale = VisualizationOptions.defaultLabelSphereScale;

	private final Color3f black = new Color3f(.0f, .0f, .0f);

	// default label color is black
	private Color3f labelColor = VisualizationOptions.labelConceptSphereColor;

	// sphere base color
	private Color3f sphereColor = null;

	// transparency of the sphere material. Default opaque.
	private float transparency = VisualizationOptions.conceptSphereTransparency;

	// Sphere radius
	private float radius = VisualizationOptions.conceptSphereRadius;

	// tells if the sphere is pickable for the user, and if it has a label
	private boolean isPickable = false, isWithLabel = false,
			hasInstances = false;

	// the sphere drawn
	private Sphere sphere = null;

	/**
	 * Void constructor. It does notingh. To draw relation, one of the
	 * createXXXSphere function must be used
	 */
	public ConceptSphere(Color3f color, String label, boolean isPickable,
			boolean hasInstances) {

		this.sphereColor = color;
		if (label != null) {
			this.sphereLabel = label;
			this.isWithLabel = true;
		} else
			this.isWithLabel = false;
		this.isPickable = isPickable;
		this.hasInstances = hasInstances;
	}

	/**
	 * Creates and adds a Sphere to a TransformGroup. The user data (which is
	 * the information retrieved by the Picker Behavior) of the 3D sphere object
	 * is set as the label string.
	 * 
	 * @return the BranchGroup containing the sphere
	 */
	public BranchGroup createColoredSphere() {
		// crea la sfera e la aggiunge al tgx
		BranchGroup bgx = new BranchGroup();
		bgx.setCapability(BranchGroup.ALLOW_DETACH);
		bgx.setCapability(TransformGroup.ALLOW_CHILDREN_READ);
		bgx.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		Material material = new Material(this.sphereColor, black,
				this.sphereColor, this.sphereColor, 10.0f);

		material.setLightingEnable(true);
		Appearance app = new Appearance();
		app.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
		app.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		app.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		app.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);

		if (this.transparency > 0) {
			TransparencyAttributes ta = new TransparencyAttributes(
					TransparencyAttributes.BLENDED, this.transparency);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
			app.setTransparencyAttributes(ta);
		}

		app.setMaterial(material);

		sphere = new Sphere(radius, Sphere.ENABLE_APPEARANCE_MODIFY
				| Sphere.GENERATE_NORMALS, 50, app);

		if (this.isPickable) {
			if (sphereLabel != null) {
				sphere.setPickable(true);
				sphere.setUserData(new String(sphereLabel));
			}
		} else
			sphere.setPickable(false);

		sphere.setCapability(Sphere.ALLOW_BOUNDS_READ);
		sphere.setCapability(Shape3D.ALLOW_LOCAL_TO_VWORLD_READ);
		bgx.addChild(sphere);

		if (this.isWithLabel)
			bgx.addChild(this.addSphereLabel());

		if (this.hasInstances) {
			// Instance aura Color
			float r = (float) Math.random();
			float g = (float) Math.random();
			float b = (float) Math.random();
			Color3f glowColor = new Color3f(r, g, b);

			ConceptSphere glowingSfera = new ConceptSphere(glowColor, null,
					false, false);
			glowingSfera.setTransparency(VisualizationOptions.sG_InstanceShellTransparency);
			glowingSfera.setRadius(this.radius* VisualizationOptions.sG_instanceShellRadiusMultiplier);

			bgx.addChild(glowingSfera.createColoredSphere());
		}

		return bgx;
	}

	/**
	 * @param transparency
	 *            The transparency to set.
	 */
	public void setTransparency(float transparency) {
		this.transparency = transparency;
	}

	/**
	 * This method creates a TransformGroup containing the label.
	 * 
	 * @return the TransformGroup containing the label
	 */
	private TransformGroup addSphereLabel() {
		// inserisce la label e la aggiunge al tgx
		String textString = sphereLabel.substring(sphereLabel.indexOf("#") + 1);
		float sl = textString.length();

		// create 3D text
		Font3D f3d = new Font3D(new Font(VisualizationOptions.sG_FontFamily,
				VisualizationOptions.sG_FontModifier, 2), new FontExtrusion());

		Point3f textPt;
		float fontHeight;
		fontHeight = (this.radius + .02f) / this.labelScale;
		
		textPt = new Point3f(-sl / 2.0f, fontHeight, 0.0f);
		Text3D txt = new Text3D(f3d, textString, textPt);

		OrientedShape3D textShape = new OrientedShape3D();
		textShape.setGeometry(txt);
		textShape.setPickable(false);

		TransformGroup objScale = new TransformGroup();
		Transform3D textMat = new Transform3D();
		textMat.setScale(this.labelScale);
		objScale.setTransform(textMat);

		Appearance apText = new Appearance();
		ColoringAttributes colorTxt = new ColoringAttributes(this.labelColor, 1);
		apText.setColoringAttributes(colorTxt);
		apText.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		apText.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		apText.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		textShape.setAppearance(apText);

		textShape.setAlignmentMode(OrientedShape3D.ROTATE_ABOUT_POINT);
		Point3f rotationPt = new Point3f(0.0f, 0f, 0.0f);
		textShape.setRotationPoint(rotationPt);

		objScale.addChild(textShape);
		return objScale;
	}

	public Appearance getSphereAppearance() {
		Appearance app = sphere.getAppearance();
		return app;
	}

	/**
	 * @param radius
	 *            The radius to set.
	 */
	public void setRadius(float radius) {
		this.radius = radius;
	}

	/**
	 * @return Returns the sphere.
	 */
	public Sphere getSphere() {
		return sphere;
	}

	/**
	 * @return Returns the radius.
	 */
	public float getRadius() {
		return radius;
	}

	/**
	 * @param labelColor
	 *            The labelColor to set.
	 */
	public void setLabelColor(Color3f labelColor) {
		this.labelColor = labelColor;
	}

	/**
	 * @param labelScale
	 *            The labelScale to set.
	 */
	public void setLabelScale(float labelScale) {
		this.labelScale = labelScale;
	}

	public void setWithLabel(boolean isWithLabel) {
		this.isWithLabel = isWithLabel;
	}
}
