package it.polito.elite.visualization.ontoviewer.tools;

import it.polito.elite.ontology.NodeTypeAbstraction;
import it.polito.elite.visualization.ontoviewer.enums.GraphicalLinkTypeEnum;
import it.polito.elite.visualization.ontoviewer.enums.LinkDirectionEnum;
import it.polito.elite.visualization.ontoviewer.enums.NodeTypesEnum;
import it.polito.elite.visualization.ontoviewer.models.RelationData;
import it.polito.elite.visualization.ontoviewer.models.RelationType;
import it.polito.elite.visualization.ontoviewer.userInterfaces.VisualizationOptions;

import java.awt.Color;
import java.awt.Font;
import java.util.HashSet;
import java.util.Hashtable;

import javax.media.j3d.Appearance;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.Font3D;
import javax.media.j3d.FontExtrusion;
import javax.media.j3d.Material;
import javax.media.j3d.OrientedShape3D;
import javax.media.j3d.Text3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.TransparencyAttributes;
import javax.vecmath.AxisAngle4f;
import javax.vecmath.Color3f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import org.j3d.renderer.java3d.geom.Torus;

import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.geometry.Cone;
import com.sun.j3d.utils.geometry.Cylinder;
import com.sun.j3d.utils.geometry.Sphere;

/**
 * This class offers an utility to draw relations as a building block for
 * scenes.It exposes a dummy constructor, a group of set methods in order to
 * explicitly set characteristics as transparency, the radius of line.. It
 * provides also a set of method createRelationXXX (as createISARelation..)
 * which return a java 3D TransformGroup that will be included by the scene
 * itself. An example of its use is: TransformGroup tgx.... Relation3DObject
 * relation = new Relation3DObject(); relation.setRadiusOfCylinderLine(.004f);
 * relation.setTransparency(.3f); BranchGroup branchGroup =
 * relation.createISARelation(relationEnd); tgx.addChild(branchGroup);
 * 
 * @author alessio.bosca@polito.it
 * 
 */
public class Relation3DObject {

	static private Hashtable<String, Color> coloredRelations = new Hashtable<String, Color>();

	private boolean isCardinalityConstrained = false, isDatatype = false;

	private final Color3f black = new Color3f(.0f, .0f, .3f);

	private Color3f relationColor = black;

	// extreme of the line
	private Vector3f relativeEndingOBJPosition, translationVector;

	private GraphicalLinkTypeEnum relationType = null;

	private LinkDirectionEnum relationDirection;

	// transparency value & label scaling factor & source elemnt size (cube side
	// or sphere radius)
	private float transparency = .0f, sourceElementSize = .0f;

	private String uriOfPickableEndingOBJ = null;

	private HashSet<RelationData> setOfRelationData;

	/**
	 * Void constructor. It does notingh. To draw relation, one of the
	 * createRelation function must be used
	 * 
	 */
	public Relation3DObject() {
		this.setOfRelationData = new HashSet<RelationData>();
	}

	/**
	 * Creates a relation with an ending 3D icon that represents the range of
	 * the relation
	 * 
	 */
	public BranchGroup createRelationWithRange(RelationData rdata,
			Vector3f relativeRangePosition, LinkDirectionEnum direction) {
		this.isCardinalityConstrained = rdata.isCardinalityConstrained();
		if (this.relationType == null)
			this.relationType = GraphicalLinkTypeEnum.withEndingObj;
		this.isDatatype = (rdata.getRelatedNode().getType() == NodeTypesEnum.DataType);
		this.relationDirection = direction;
		this.setOfRelationData.add(rdata);
		this.relativeEndingOBJPosition = relativeRangePosition;
		return this.drawRelationWithPlaceHolder(rdata);
	}

	/**
	 * Creates a relation with an ending 3D icon that represents the range of
	 * the relation
	 * 
	 * @return a BranchGroup containing the line and the 3d icon complete with
	 *         labels
	 */
	public BranchGroup createRelationWithRange(String relation,
			NodeTypeAbstraction range, Vector3f relativeRangePosition,
			LinkDirectionEnum direction) {
		if (this.relationType == null)
			this.relationType = GraphicalLinkTypeEnum.withEndingObj;
		this.relationDirection = direction;
		this.isDatatype = false;
		if (relation != null)
			this.setOfRelationData.add(new RelationData(relation, direction, new RelationType() ));
		this.relativeEndingOBJPosition = relativeRangePosition;
		return this.drawRelationWithPlaceHolder(range);
	}

	/**
	 * Creates a relation with an ending 3D icon that represents the Set of the
	 * ranges
	 * 
	 * @param ranges
	 *            The set of range elements
	 * @param relation
	 *            URI of the relation
	 * @param relativeRangePosition
	 *            The position of the ending object
	 * @param direction
	 *            The direction of the relation (Must be LinkTypeEnum.in or
	 *            LinkTypeEnum.out)
	 * @return a BranchGroup containing the line and the 3d icon complete with
	 *         labels
	 */
	public BranchGroup createRelationWithSetOfElements(String relation,
			String groupingID, HashSet<String> ranges,
			Vector3f relativeRangePosition, LinkDirectionEnum direction,
			boolean isConcept) {
		if (this.relationType == null)
			this.relationType = GraphicalLinkTypeEnum.withEndingObj;
		this.relationDirection = direction;
		this.relativeEndingOBJPosition = relativeRangePosition;
		if (relation != null) {
			this.setOfRelationData.add(new RelationData(relation, direction, new RelationType()));
		}
		return this
				.drawRelationWithNamedEntities(groupingID, ranges, isConcept);
	}


	public BranchGroup createSideRelations(
			Hashtable<String, HashSet<NodeTypeAbstraction>> relationsNotOnScreen,
			Vector3f relativeRangePosition, LinkDirectionEnum direction) {
		if (this.relationType == null)
			this.relationType = GraphicalLinkTypeEnum.withEndingObj;
		this.relationDirection = direction;
		this.relativeEndingOBJPosition = relativeRangePosition;
		for (String relation : relationsNotOnScreen.keySet()) {
			this.setOfRelationData.add(new RelationData(relation, direction, new RelationType()));
		}
		if (relationsNotOnScreen.size() == 1
				&& relationsNotOnScreen.values().iterator().next().size() == 1) {// simplest
																					// situation..
			NodeTypeAbstraction range = relationsNotOnScreen.values()
					.iterator().next().iterator().next();
			return this.drawRelationWithPlaceHolder(range);
		} else {
			HashSet<String> ranges = new HashSet<String>();
			for (HashSet<NodeTypeAbstraction> set : relationsNotOnScreen
					.values()) {
				for (NodeTypeAbstraction node : set) {
					ranges.add(node.toString());
				}
			}
			return this.drawRelationWithNamedEntities(null, ranges, true);
		}
	}

	/**
	 * Draws a plain ISA relation to the specified relative position
	 * 
	 * @param relativeEndingPosition
	 *            The position has to be equal to the difference between the
	 *            parent and the son absolute positions
	 * @return a BranchGroup containing the relation line
	 */
	public BranchGroup createISARelation(Vector3f relativeEndingPosition) {
		this.isCardinalityConstrained = false;
		this.relationType = GraphicalLinkTypeEnum.isA;
		this.relationDirection = LinkDirectionEnum.out;
		this.relativeEndingOBJPosition = relativeEndingPosition;
		return this.drawArrowedLine();
	}

	/**
	 * It draws a plain (not an ISA) relation
	 * 
	 * @return a BranchGroup containing the relation line
	 */
	public BranchGroup createStraightRelation(HashSet<RelationData> relations,
			Vector3f relativeEndingPosition, float sourceSphereRadius,
			LinkDirectionEnum dir) {
		this.relationType = GraphicalLinkTypeEnum.plain;
		this.sourceElementSize = sourceSphereRadius;
		this.relationDirection = dir;
		this.setOfRelationData = relations;
		this.relativeEndingOBJPosition = relativeEndingPosition;
		return this.drawArrowedLine();
	}

	public BranchGroup createInstanceStraightRelation(Hashtable<String, RelationType> relations, Vector3f relativeEndingPosition,
			float sourceSphereRadius, LinkDirectionEnum dir) {
		this.relationType = GraphicalLinkTypeEnum.plain;
		this.sourceElementSize = sourceSphereRadius;
		this.relationDirection = dir;
		for (String str : relations.keySet()) {
			this.setOfRelationData.add(new RelationData(str, dir, relations.get(str)));
		}
		this.relativeEndingOBJPosition = relativeEndingPosition;
		return this.drawArrowedLine();
	}

	public BranchGroup createSelfLoopRelation(HashSet<RelationData> relations,
			float sourceSphereRadius, AxisAngle4f angle) {
		this.relationType = GraphicalLinkTypeEnum.selfLoop;
		this.sourceElementSize = sourceSphereRadius;
		this.relationDirection = LinkDirectionEnum.out;
		this.setOfRelationData = relations;
		return drawArrowedLoop(angle);
	}

	public BranchGroup createInstanceSelfLoopRelation(	Hashtable<String, RelationType> relations, float sourceSphereRadius,
			AxisAngle4f angle) {
		this.isCardinalityConstrained = isCardinalityConstrained;
		this.relationType = GraphicalLinkTypeEnum.selfLoop;
		this.sourceElementSize = sourceSphereRadius;
		this.relationDirection = LinkDirectionEnum.out;
		for (String relUri : relations.keySet()) {
			this.setOfRelationData.add(new RelationData(relUri, LinkDirectionEnum.out, relations.get(relUri)));
		}
		return drawArrowedLoop(angle);
	}

	public BranchGroup createCurvedArrowRelation(
			HashSet<RelationData> relations, Vector3f relativeEndingPosition,
			float curvefactor, LinkDirectionEnum dir) {
		this.relationType = GraphicalLinkTypeEnum.plain;
		this.sourceElementSize = 0;
		this.relationDirection = dir;
		this.setOfRelationData = relations;
		this.relativeEndingOBJPosition = relativeEndingPosition;
		return drawCurvedArrow(curvefactor);
	}

	public BranchGroup createInstanceCurvedArrowRelation(
			Hashtable<String, RelationType> relations, Vector3f relativeEndingPosition,
			float curvefactor, LinkDirectionEnum dir) {
		this.relationType = GraphicalLinkTypeEnum.plain;
		this.sourceElementSize = 0;
		this.relationDirection = dir;
		for (String relUri : relations.keySet()) {
			this.setOfRelationData.add(new RelationData(relUri, LinkDirectionEnum.out, relations.get(relUri)));
		}
		this.relativeEndingOBJPosition = relativeEndingPosition;
		return drawCurvedArrow(curvefactor);
	}

	private BranchGroup drawCurvedArrow(float scale) {

		BranchGroup cylBg = this.createRelationBG();
		Appearance appearance = this.createLineAppearance();

		Vector3f normal;
		if (this.relativeEndingOBJPosition.x == 0f
				&& this.relativeEndingOBJPosition.z == 0f)
			normal = new Vector3f(1f, 0f, 0f);
		else {
			Vector3f base0 = new Vector3f(this.relativeEndingOBJPosition.x, 0f,
					this.relativeEndingOBJPosition.z);
			Vector3f base = new Vector3f();
			base.cross(this.relativeEndingOBJPosition, base0);
			base.normalize();
			normal = new Vector3f();
			normal.cross(this.relativeEndingOBJPosition, base);
			normal.normalize();

			// Choose the normal directed towards the extern of the "cone"
			// created by a concept and his "sons"
			float squareDistPos = (normal.x - base0.x) * (normal.x - base0.x)
					+ (normal.z - base0.z) * (normal.z - base0.z);
			float squareDistNeg = (-normal.x - base0.x) * (-normal.x - base0.x)
					+ (-normal.z - base0.z) * (-normal.z - base0.z);
			if (squareDistNeg > squareDistPos
					&& this.relativeEndingOBJPosition.y > 0)
				normal.negate();
		}
		normal.scale(this.relativeEndingOBJPosition.length());
		normal.scale(scale);

		// calculate the trajectory as cylinder segments
		Vector3f last = new Vector3f(0f, 0f, 0f);
		int tot = 10;
		float pace = (float) (1f / ((float) tot));
		int coneSlot = tot / 2;
		for (int i = 1; i <= tot; i++) {
			Vector3f pos = new Vector3f(this.relativeEndingOBJPosition);
			pos.scale(pace * i);
			Vector3f nbase = new Vector3f(normal);
			float nscale = (float) Math.sin(Math.PI * pace * i);
			nbase.scale(nscale);
			pos.add(nbase);
			Vector3f partial = new Vector3f(pos.x - last.x, pos.y - last.y,
					pos.z - last.z);

			// Calculate translation
			Transform3D translation = new Transform3D();
			translation.setTranslation(last);
			TransformGroup tg0 = new TransformGroup(translation);

			// calculate line orientation creating transMatrix

			float cylinderHeight = partial.length();
			Transform3D rot = this.computeOrientation(partial, 0.5f, false,
					false);
			// create the transform group
			TransformGroup tg = new TransformGroup(rot);

			// Creates the cylinder-line which has its center in
			// the middle between the 2 endpoints
			Cylinder cyl = new Cylinder(VisualizationOptions.sR_cylinderRadius,
					cylinderHeight, appearance);
			cyl.setPickable(false);
			tg.addChild(cyl);
			tg0.addChild(tg);

			if (i == coneSlot) {
				boolean invertDirection = false;
				if (this.relationDirection == LinkDirectionEnum.in)
					invertDirection = true;

				Transform3D rotCone = this.computeOrientation(partial, 0.5f,
						invertDirection, true);

				TransformGroup coneTr = this.drawArrowHeadCone(rotCone,
						setOfRelationData);
				tg0.addChild(coneTr);
			}
			cylBg.addChild(tg0);// adds the cylinder

			last = pos;
		}
		return cylBg;
	}

	private BranchGroup drawArrowedLoop(AxisAngle4f angle) {

		BranchGroup bg = this.createRelationBG();
		Appearance appearance = this.createLineAppearance();

		float outerRadius = VisualizationOptions.sR_selfLoopRadiusMultiplier
				* this.sourceElementSize;
		float innerRadius = VisualizationOptions.sR_cylinderRadius;

		// Create the orientation rotation
		Transform3D rot0 = new Transform3D();
		if (angle != null)
			rot0.setRotation(angle);
		TransformGroup tg0;
		try {
			tg0 = new TransformGroup(rot0);
		} catch (javax.media.j3d.BadTransformException e) {
			tg0 = new TransformGroup(new Transform3D());
		}

		// create the transform group
		Transform3D rot = new Transform3D();
		rot.rotX(Math.PI / 2);

		rot.setTranslation(new Vector3f((outerRadius), outerRadius
				* VisualizationOptions.sR_selfLoopverticalMultiplier, 0f));
		TransformGroup tg = new TransformGroup(rot);

		Torus loop = new Torus(innerRadius, outerRadius, appearance);
		loop.setPickable(false);

		Transform3D coneRot = new Transform3D();

		coneRot.rotZ(Math.PI / 2);
		coneRot.setTranslation(new Vector3f(0f, 0f, -outerRadius));

		TransformGroup coneTrans = this.drawArrowHeadCone(coneRot,
				setOfRelationData);
		tg.addChild(coneTrans);// adds the cone

		tg.addChild(loop);
		tg0.addChild(tg);
		bg.addChild(tg0);
		return bg;
	}

	/**
	 * Draws the line (cylinder + cone)
	 * 
	 * @return a BranchGroup containing the line
	 */
	private BranchGroup drawArrowedLine() {

		BranchGroup cylBg = this.createRelationBG();
		Appearance appearance = this.createLineAppearance();

		// represents the position (between 0 origin and 1 end position) where
		// an object should be placed
		float scale = .5f;

		// calculate line orientation creating transMatrix

		float lineLength = relativeEndingOBJPosition.length();
		Transform3D rot = this.computeOrientation(relativeEndingOBJPosition,
				scale, false, false);

		// create the transform group
		TransformGroup tg = new TransformGroup();
		tg = new TransformGroup(rot);

		// Creates the cylinder-line which has its center in
		// the middle between the 2 endpoints
		if (!this.isCardinalityConstrained) {
			Cylinder cyl = new Cylinder(VisualizationOptions.sR_cylinderRadius,
					lineLength, appearance);
			cyl.setPickable(false);
			tg.addChild(cyl);
		} else {
			Box cube = new Box(VisualizationOptions.sR_cylinderRadius
					* VisualizationOptions.sR_boxedArrowMultiplier,
					lineLength * .5f, VisualizationOptions.sR_cylinderRadius
							* VisualizationOptions.sR_boxedArrowMultiplier,
					Box.ENABLE_APPEARANCE_MODIFY | Box.GENERATE_NORMALS,
					appearance);
			cube.setPickable(false);
			tg.addChild(cube);
		}

		// Creates a cone-arrow with an optional label at ~2/3 of the distance
		// between ends if outGoing and ~1/3 if incoming
		scale = this.computeConePositionScale();
		boolean invert;
		if ((this.relationDirection == LinkDirectionEnum.in))// &&
																// this.relationType
																// ==
																// GraphicalLinkTypeEnum.withEndingObj
			invert = true;
		else
			invert = false;

		rot = this.computeOrientation(relativeEndingOBJPosition, scale, invert,
				true);

		TransformGroup coneTr = this.drawArrowHeadCone(rot, setOfRelationData);
		cylBg.addChild(coneTr);// adds the cone

		cylBg.addChild(tg);// adds the cylinder
		return cylBg;

	}

	private TransformGroup drawArrowHeadCone(Transform3D where,
			HashSet<RelationData> relLabelSet) {

		Appearance appearance = this.createArrowHeadAppearance((relLabelSet
				.size() > 1));
		TransformGroup coneTrans = new TransformGroup(where);

		float coneRadius = VisualizationOptions.sR_coneRadius
				* (1 + relLabelSet.size()
						* VisualizationOptions.sR_MultiRelationMultiplier);
		if (coneRadius > VisualizationOptions.sR_MultiRelationRadiusLimit)
			coneRadius = VisualizationOptions.sR_MultiRelationRadiusLimit;
		float coneHeight = VisualizationOptions.sR_coneHeight
				* (1 + relLabelSet.size()
						* VisualizationOptions.sR_MultiRelationMultiplier);
		if (coneHeight > VisualizationOptions.sR_MultiRelationHeightLimit)
			coneHeight = VisualizationOptions.sR_MultiRelationHeightLimit;
		Cone cono = new Cone(coneRadius, coneHeight, Cone.GENERATE_NORMALS
				| Cone.GENERATE_TEXTURE_COORDS, appearance);

		if (relLabelSet.size() > 0) {// adds the label
			cono.setUserData(this.setOfRelationData);
			coneTrans.setPickable(true);
		} else {
			coneTrans.setPickable(false);
		}
		coneTrans.addChild(cono);// adding the arrow cone

		if (relLabelSet.size() == 1
				&& VisualizationOptions.sR_DisplayRelationsLabels) {
			RelationData reldata = relLabelSet.iterator().next();
			BranchGroup relBg = this.drawLabel(reldata.getRelationName());

			// Set the relative translatio of the label w.r.t the cone
			Transform3D relTr = new Transform3D();
			if (this.relationType == GraphicalLinkTypeEnum.selfLoop) {
				relTr
						.setTranslation(new Vector3f(
								0f,
								0f,
								-(coneRadius / 2 + VisualizationOptions.sR_LabelPositionOffset)));
			} else {
				Vector3f translation = new Vector3f(this.translationVector);
				translation.y += (coneRadius + VisualizationOptions.sR_LabelPositionOffset);
				relTr.setTranslation(translation);
				TransformGroup all = new TransformGroup();
				TransformGroup relTg = new TransformGroup(relTr);
				relTg.addChild(relBg);
				all.addChild(relTg);
				all.addChild(coneTrans);
				return all;
			}
			TransformGroup relTg = new TransformGroup(relTr);
			relTg.addChild(relBg);

			coneTrans.addChild(relTg);
		}

		return coneTrans;
	}

	/**
	 * Creates a 3D label object
	 * 
	 * @param relationLabel The label
	 * @return a TransformGroup containing the 3D object containing the label
	 */
	private BranchGroup drawLabel(String relationLabel) {
		BranchGroup bg = this.createRelationBG();

		if (this.setOfRelationData == null)
			return null;
		/* scrivi la label della relazione */
		// create 3D text
		Font3D f3d2 = new Font3D(new Font(VisualizationOptions.sG_FontFamily,
				VisualizationOptions.fontModifier, 2), new FontExtrusion());
		Point3f textPt2 = new Point3f(-relationLabel.length() / 2.0f, 1.2f,
				0.0f);
		Text3D txt2 = new Text3D(f3d2, relationLabel, textPt2);

		OrientedShape3D textShape2 = new OrientedShape3D();
		textShape2.setGeometry(txt2);
		textShape2.setPickable(false);

		TransformGroup objScale2 = new TransformGroup();
		Transform3D textMat2 = new Transform3D();
		textMat2.setScale(VisualizationOptions.sR_RelationsLabeScale);

		objScale2.setTransform(textMat2);
		Appearance apText2 = new Appearance();
		ColoringAttributes colorTxt2 = new ColoringAttributes(
				this.relationColor, 1);
		apText2.setColoringAttributes(colorTxt2);
		apText2.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		apText2.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		apText2.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		textShape2.setAppearance(apText2);

		textShape2.setAlignmentMode(OrientedShape3D.ROTATE_ABOUT_POINT);
		Point3f rotationPt2 = new Point3f(0.0f, 0.0f, 0.0f);
		textShape2.setRotationPoint(rotationPt2);

		textShape2.setPickable(false);
		objScale2.addChild(textShape2);
		bg.addChild(objScale2);

		return bg;
	}

	/**
	 * Metodhs that adjusts graphics setting for the appearance of the
	 * components and returns a proper appearance
	 * 
	 */
	private Appearance createLineAppearance() {
		Appearance apparenza = new Appearance();
		Material material;

		if (this.relationType == GraphicalLinkTypeEnum.isA) {
			relationColor = VisualizationOptions.colorLine;
		} else {
			boolean hasCustomColor=false;
			if (this.setOfRelationData.size() == 1) {
				String relationURI = ((RelationData) this.setOfRelationData
						.iterator().next()).getRelationName();
				if (Relation3DObject.coloredRelations.containsKey(relationURI)) {
					Color custumCol = Relation3DObject.coloredRelations
							.get(relationURI);
					relationColor = new Color3f(custumCol);
					hasCustomColor=true;
				}
			}
			if(!hasCustomColor){
				if (this.relationDirection == LinkDirectionEnum.in)
					relationColor = VisualizationOptions.sR_IncomingRelationColor;
				else if (this.isDatatype)
					relationColor = VisualizationOptions.sR_DatatypeRelationColor;
				else
					relationColor = VisualizationOptions.sR_OutgoingRelationColor;
			}
				
		}

		material = new Material(relationColor, black, relationColor,
				relationColor, 100.0f);
		// material.setLightingEnable(true);
		apparenza.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		apparenza.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		apparenza.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		apparenza.setMaterial(material);

		if (this.transparency > 0) {
			TransparencyAttributes ta = new TransparencyAttributes();
			ta.setTransparency(this.transparency);
			ta.setTransparencyMode(TransparencyAttributes.BLENDED);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
			apparenza.setTransparencyAttributes(ta);
		}

		return apparenza;
	}

	/**
	 * Metodhs that adjusts graphics setting for the appearance of the
	 * components and returns a proper appearance
	 * 
	 */
	private Appearance createArrowHeadAppearance(boolean isMultipleRelation) {
		Appearance apparenza = new Appearance();
		Material material;

		if (isMultipleRelation) {
			relationColor = new Color3f(
					VisualizationOptions.sR_MultiRelationArrowHeadColor);
		}

		material = new Material(relationColor, black, relationColor,
				relationColor, 10.0f);
		material.setLightingEnable(true);
		apparenza.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		apparenza.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		apparenza.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		apparenza.setMaterial(material);

		if (transparency > 0) {
			TransparencyAttributes ta = new TransparencyAttributes();
			ta.setTransparency(transparency
					* VisualizationOptions.sR_ArrowHeadTransparencyMultiplier);
			ta.setTransparencyMode(TransparencyAttributes.BLENDED);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
			apparenza.setTransparencyAttributes(ta);
		}

		return apparenza;
	}

	private Appearance createRangeItemAppearance(Color3f itemColor) {
		return this.createRangeItemAppearance(itemColor, 0);
	}

	/**
	 * Metodhs that adjusts graphics setting for the appearance of the
	 * components and returns a proper appearance
	 */
	private Appearance createRangeItemAppearance(Color3f itemColor,
			float transparence) {
		Appearance apparenza = new Appearance();
		Material material = new Material(itemColor, black, itemColor,
				itemColor, 10.0f);
		material.setLightingEnable(true);
		apparenza.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		apparenza.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		apparenza.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		apparenza.setMaterial(material);

		if (transparence > 0) {
			TransparencyAttributes ta = new TransparencyAttributes();
			ta.setTransparency(transparence);
			ta.setTransparencyMode(TransparencyAttributes.BLENDED);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
			ta.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
			apparenza.setTransparencyAttributes(ta);
		}
		return apparenza;
	}

	private Transform3D computeOrientation(Vector3f finalPos, float scale,
			boolean invertDirection, boolean isCone) {

		// Define end points as new vectors
		Vector3f fromDir = new Vector3f(0f, 1f, 0f);
		Vector3f toDir = new Vector3f(finalPos);
		toDir.normalize();

		Transform3D resultantTransform = new Transform3D();

		translationVector = new Vector3f(finalPos);
		if (isCone) {
			// Calculate proper scale with respect to the size of elements sizes
			float endingObjElmSize;
			if (this.relationType != GraphicalLinkTypeEnum.withEndingObj)
				endingObjElmSize = this.sourceElementSize;
			else
				endingObjElmSize = VisualizationOptions.sR_sphereRadius;
			float offsetFromSphere = (translationVector.length()
					- this.sourceElementSize - endingObjElmSize)
					* scale;
			float newScale = (this.sourceElementSize + offsetFromSphere)
					/ translationVector.length();

			// along y axis and facing downward, out direction
			if (toDir.equals(new Vector3f(0f, -1f, 0f))
					&& this.relationDirection == LinkDirectionEnum.out) {
				resultantTransform.rotX(Math.PI);
			}
			// along y axis and facing upward, in direction
			else if (toDir.equals(new Vector3f(0f, 1f, 0f))
					&& this.relationDirection == LinkDirectionEnum.in) {
				resultantTransform.rotX(Math.PI);
			}

			translationVector.scale(newScale);

		} else
			translationVector.scale(scale);

		resultantTransform.setTranslation(translationVector);

		if (toDir.equals(fromDir) || toDir.equals(new Vector3f(0f, -1f, 0f))) {
			return resultantTransform;
		}

		if (invertDirection) {
			toDir.negate();
		}

		// Compute the value of the angle between them (no spatial orientation
		// info though)
		float rotInRadians = fromDir.angle(toDir);

		// Compute an orthagonal vector to later define the rotation axis

		Vector3f orthagonalAxis = new Vector3f();
		orthagonalAxis.cross(fromDir, toDir);
		orthagonalAxis.normalize();

		// Create new rotation matrix
		AxisAngle4f rotAxis = new AxisAngle4f(orthagonalAxis, rotInRadians);

		resultantTransform.setRotation(rotAxis);

		return resultantTransform;
	}

	/**
	 * represents the position (between 0 origin and 1 end position) where the
	 * cone should be placed
	 * 
	 * @return the correct scale for the cone in the different relations
	 *         templates
	 */
	private float computeConePositionScale() {
		float scale = 0f;
		if (this.relationType == GraphicalLinkTypeEnum.isA)
			scale = .75f;
		else if (this.relationType == GraphicalLinkTypeEnum.withEndingObj) {
			scale = .7f;
		} else {// straight plain relation
			scale = .35f;
		}
		return scale;
	}

	/**
	 * Creates a proper BranchGroup
	 * 
	 * @return a new BranchGroup
	 */
	private BranchGroup createRelationBG() {
		BranchGroup bgx = new BranchGroup();
		bgx.setCapability(BranchGroup.ALLOW_DETACH);
		bgx.setCapability(TransformGroup.ALLOW_CHILDREN_READ);
		bgx.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		bgx.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
		return bgx;
	}

	/**
	 * Creates a proper TransformGroup
	 * 
	 * @return a new TransformGroup
	 */
	private TransformGroup createRelationTG() {
		TransformGroup tgx = new TransformGroup();
		tgx.setCapability(TransformGroup.ALLOW_CHILDREN_READ);
		tgx.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		tgx.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
		return tgx;
	}

	private BranchGroup drawRelationWithPlaceHolder(RelationData rdata) {
		NodeTypeAbstraction myRange = rdata.getRelatedNode();
		return drawRelationWithPlaceHolder(myRange);
	}

	private BranchGroup drawRelationWithNamedEntities(String groupingID,
			HashSet<String> ranges, boolean isConcept) {
		BranchGroup bgx = this.createRelationBG();

		// Drawing the arrowed Line
		TransformGroup tgxLine = this.createRelationTG();
		BranchGroup lineRel = this.drawArrowedLine();
		tgxLine.addChild(lineRel);
		bgx.addChild(tgxLine);

		// Drawing the range
		Vector3f position = this.relativeEndingOBJPosition;

		TransformGroup tgx = this.createRelationTG();
		Transform3D transf = new Transform3D();
		transf.setTranslation(position);
		tgx.setTransform(transf);

		if (groupingID == null)
			groupingID = "";
		StringBuffer buff = new StringBuffer(groupingID + " (");
		for (String str : ranges) {
			buff.append(str + ",");
		}
		this.uriOfPickableEndingOBJ = buff.substring(0, buff.length() - 1)
				+ ")";
		Color3f rangeLabelColor = new Color3f(.0f, .5f, .5f);

		Appearance apparenza1 = this.createRangeItemAppearance(new Color3f(.0f,
				.2f, .6f), .3f);// BLU
		Appearance apparenza2 = this.createRangeItemAppearance(new Color3f(1f,
				.1f, .1f));// RED
		Appearance apparenza3 = this.createRangeItemAppearance(new Color3f(.0f,
				.8f, .2f));// GREEN
		Appearance apparenza4 = this.createRangeItemAppearance(new Color3f(.6f,
				.0f, .6f));// VIOLET
		Appearance apparenza5 = this.createRangeItemAppearance(new Color3f(.5f,
				.25f, .0f));// BROWN
		Appearance apparenza6 = this.createRangeItemAppearance(new Color3f(.5f,
				.1f, .0f));// LIGHT GREEN

		// creating the range 3D Icon
		float side = VisualizationOptions.sR_cubeSide * .3f;// cube side
		float radius = VisualizationOptions.sR_sphereRadius * 1.3f;// sphere
																	// radius
		BranchGroup bgIcon3d = this.createRelationBG();
		// Sphere 1
		TransformGroup tg1 = new TransformGroup();
		Sphere sphere = new Sphere(radius, Sphere.ENABLE_APPEARANCE_MODIFY
				| Sphere.GENERATE_NORMALS, 50, apparenza1);
		sphere.setPickable(true);
		sphere.setUserData(this.uriOfPickableEndingOBJ);
		tg1.addChild(sphere);

		// Sphere 1 up
		Transform3D trans2 = this.computeOrientation(new Vector3f(0f,
				radius * .7f, 0f), 1.3f, false, false);
		TransformGroup tg2 = new TransformGroup(trans2);
		if (isConcept) {
			Sphere sphere1 = new Sphere(side, Sphere.ENABLE_APPEARANCE_MODIFY
					| Sphere.GENERATE_NORMALS, 50, apparenza2);
			sphere1.setPickable(false);
			tg2.addChild(sphere1);
		} else {
			Box cube1 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
					| Box.GENERATE_NORMALS, apparenza2);
			cube1.setPickable(false);
			tg2.addChild(cube1);
		}
		// Sphere 2
		Transform3D trans3 = this.computeOrientation(new Vector3f(radius * .7f,
				0, radius * .4f), 1.3f, false, false);
		TransformGroup tg3 = new TransformGroup(trans3);
		if (isConcept) {
			Sphere sphere1 = new Sphere(side, Sphere.ENABLE_APPEARANCE_MODIFY
					| Sphere.GENERATE_NORMALS, 50, apparenza3);
			sphere1.setPickable(false);
			tg3.addChild(sphere1);
		} else {
			Box cube1 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
					| Box.GENERATE_NORMALS, apparenza3);
			cube1.setPickable(false);
			tg3.addChild(cube1);
		}
		// sphere 3
		Transform3D trans4 = this.computeOrientation(new Vector3f(
				-radius * .7f, 0, radius * .4f), 1.3f, false, false);
		TransformGroup tg4 = new TransformGroup(trans4);
		if (isConcept) {
			Sphere sphere1 = new Sphere(side, Sphere.ENABLE_APPEARANCE_MODIFY
					| Sphere.GENERATE_NORMALS, 50, apparenza4);
			sphere1.setPickable(false);
			tg4.addChild(sphere1);
		} else {
			Box cube1 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
					| Box.GENERATE_NORMALS, apparenza4);
			cube1.setPickable(false);
			tg4.addChild(cube1);
		}
		// sphere 4 DOWN
		Transform3D trans5 = this.computeOrientation(new Vector3f(0,
				-radius * .7f, 0), 1.3f, false, false);
		TransformGroup tg5 = new TransformGroup(trans5);
		if (isConcept) {
			Sphere sphere1 = new Sphere(side, Sphere.ENABLE_APPEARANCE_MODIFY
					| Sphere.GENERATE_NORMALS, 50, apparenza5);
			sphere1.setPickable(false);
			tg5.addChild(sphere1);
		} else {
			Box cube1 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
					| Box.GENERATE_NORMALS, apparenza5);
			cube1.setPickable(false);
			tg5.addChild(cube1);
		}
		// sphere 5 CIRC
		Transform3D trans6 = this.computeOrientation(new Vector3f(0f, 0,
				-radius * .7f), 1.3f, false, false);
		TransformGroup tg6 = new TransformGroup(trans6);
		if (isConcept) {
			Sphere sphere1 = new Sphere(side, Sphere.ENABLE_APPEARANCE_MODIFY
					| Sphere.GENERATE_NORMALS, 50, apparenza6);
			sphere1.setPickable(false);
			tg6.addChild(sphere1);
		} else {
			Box cube1 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
					| Box.GENERATE_NORMALS, apparenza6);
			cube1.setPickable(false);
			tg6.addChild(cube1);
		}

		bgIcon3d.addChild(tg1);
		bgIcon3d.addChild(tg2);
		bgIcon3d.addChild(tg3);
		bgIcon3d.addChild(tg4);
		bgIcon3d.addChild(tg5);
		bgIcon3d.addChild(tg6);

		tgx.addChild(bgIcon3d);
		if (groupingID != null)
			this.drawRangeLabel(tgx, groupingID, this.black);

		bgx.addChild(tgx);
		return bgx;
	}

	/**
	 * Draws a relation (arrowed line) and an ending Pickable object
	 * 
	 * @return a new BranchGroup containing the relation and the sphere
	 */
	private BranchGroup drawRelationWithPlaceHolder(NodeTypeAbstraction myRange) {
		BranchGroup bgx = this.createRelationBG();

		// Drawing the arrowed Line
		TransformGroup tgxLine = this.createRelationTG();
		BranchGroup lineRel = this.drawArrowedLine();
		tgxLine.addChild(lineRel);
		bgx.addChild(tgxLine);

		// Drawing the range
		Vector3f position = this.relativeEndingOBJPosition;

		TransformGroup tgx = this.createRelationTG();
		Transform3D transf = new Transform3D();
		transf.setTranslation(position);
		tgx.setTransform(transf);

		this.uriOfPickableEndingOBJ = myRange.toString();
		Color3f rangeLabelColor = new Color3f(.0f, .0f, .3f);

		switch (myRange.getType()) {
		case NamedClass:
			this.addNamedConceptRange(tgx);
			break;
		case NamedInstance:
			this.addNamedIndividualRange(tgx);
			break;
		case DataType:
		case Literal:
			this.addDataTypeRange(tgx);
			break;
		// Restrictions
		case QualifiedCardinality:
		case QualifiedMinCardinality:
		case QualifiedMaxCardinality:
		case Cardinality:
		case MaxCardinality:
		case MinCardinality:
		case ExistentialQuantifier:
		case UniversalQuantifier:
			this.addRestrictionRange(tgx);
			break;
		case Union:
			this.addUnionRange(tgx);
			break;
		case Intersection:
			this.addIntersectionRange(tgx);
			break;
		case ComplementOf:
			this.addComplementOf(tgx);
			break;
		case EnumeratedClass:
			this.addEnumerationRange(tgx);
			break;
		default:
		}

		bgx.addChild(tgx);
		return bgx;
	}

	private void drawRangeLabel(TransformGroup tgx, String rangeLAbel,
			Color3f rangeLabelColor) {
		// Label del concetto selezionato. create 3D text
		Font3D f3d = new Font3D(new Font("Dialog", Font.PLAIN, 2),
				new FontExtrusion());
		Point3f textPt = new Point3f(-rangeLAbel.length() / 2.0f, -5.5f, 0.0f);

		Text3D txt = new Text3D(f3d, rangeLAbel, textPt);

		OrientedShape3D textShape = new OrientedShape3D();
		textShape.setGeometry(txt);
		textShape.setPickable(false);

		TransformGroup objScale = new TransformGroup();
		Transform3D textMat = new Transform3D();
		textMat.setScale(VisualizationOptions.sR_placeHolderLabelScale);
		objScale.setTransform(textMat);

		Appearance apText = new Appearance();
		ColoringAttributes colorTxt = new ColoringAttributes(rangeLabelColor, 1);
		apText.setColoringAttributes(colorTxt);
		apText.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		apText.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		apText.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		textShape.setAppearance(apText);

		textShape.setAlignmentMode(OrientedShape3D.ROTATE_ABOUT_POINT);
		Point3f rotationPt = new Point3f(0.0f, 0.0f, 0.0f);

		textShape.setRotationPoint(rotationPt);

		objScale.addChild(textShape);
		tgx.addChild(objScale);
	}


	/*
	 * private Vector3f drawRangeLineProsecution(BranchGroup bgx, Vector3f
	 * position, int rangeNumber){ // Creating the prosecution of the rel range
	 * ->-A -B.. Vector3f posRel=new Vector3f(position); posRel.scale(.35f);
	 * Vector3f oldPos=new Vector3f(position); oldPos.scale( 1 + (
	 * (rangeNumber-1) * .35f ) ); //Updates the position Vector3f
	 * newPosition=new Vector3f(position); newPosition.scale( 1 + (
	 * (rangeNumber) * .35f ) );
	 *  // calculate line orientation creating transMatrix float cylinderHeight =
	 * posRel.length(); Transform3D rot = this.computeOrientation(posRel,.5f,
	 * false, false);
	 *  // create the transform group TransformGroup tg = new
	 * TransformGroup(rot); // Creates the cylinder-line which has its center in //
	 * the middle between the 2 endpoints Cylinder cyl = new
	 * Cylinder(VisualizationOptions.sR_cylinderRadius, cylinderHeight,
	 * this.createLineAppearance()); cyl.setPickable(false); tg.addChild(cyl);
	 *  // Adding it tho the scene Transform3D trasl = new Transform3D();
	 * trasl.setTranslation(oldPos); // create the transform group
	 * TransformGroup tg0 = new TransformGroup(trasl); tg0.addChild(tg);
	 * bgx.addChild(tg0);
	 * 
	 * return newPosition; }
	 */

	private void addNamedConceptRange(TransformGroup tgx) {
		// creating Range Appearance
		Color3f itemColor;
		if (this.relationDirection == LinkDirectionEnum.out)
			itemColor = new Color3f(.5f, .0f, .0f);// RED
		else
			itemColor = new Color3f(.0f, .5f, .0f);// GREEN
		Appearance apparenza = this.createRangeItemAppearance(itemColor);

		// creating the range 3D Icon
		Sphere sphere = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza);

		sphere.setPickable(true);
		sphere.setUserData(this.uriOfPickableEndingOBJ);
		sphere.setCapability(Sphere.ALLOW_BOUNDS_READ);

		tgx.addChild(sphere);

		this.drawRangeLabel(tgx, this.uriOfPickableEndingOBJ, itemColor);
	}

	private void addNamedIndividualRange(TransformGroup tgx) {
		// creating Range Appearance
		Color3f itemColor;
		if (this.relationDirection == LinkDirectionEnum.out)
			itemColor = new Color3f(.5f, .0f, .0f);// RED
		else
			itemColor = new Color3f(.0f, .5f, .0f);// GREEN
		Appearance apparenza = this.createRangeItemAppearance(itemColor);

		// creating the range 3D Icon
		Box cube = new Box(VisualizationOptions.sR_cubeSide,
				VisualizationOptions.sR_cubeSide,
				VisualizationOptions.sR_cubeSide, Box.ENABLE_APPEARANCE_MODIFY
						| Box.GENERATE_NORMALS, apparenza);

		cube.setPickable(true);
		cube.setUserData(this.uriOfPickableEndingOBJ);
		cube.setCapability(Sphere.ALLOW_BOUNDS_READ);

		tgx.addChild(cube);
		this.drawRangeLabel(tgx, this.uriOfPickableEndingOBJ, itemColor);
	}

	private void addDataTypeRange(TransformGroup tgx) {
		// creating Range Appearance
		Color3f dataTypeColor = new Color3f(.0f, .0f, .5f);// BLUE
		Appearance apparenza = this
				.createRangeItemAppearance(this.relationColor);

		// creating the range 3D Icon
		Cylinder cyl = new Cylinder(
				VisualizationOptions.sR_dataTypeCylinderRadius,
				VisualizationOptions.sR_dataTypeCylinderHeight, apparenza);

		cyl.setPickable(true);
		cyl.setUserData(this.uriOfPickableEndingOBJ);
		cyl.setCapability(Sphere.ALLOW_BOUNDS_READ);

		tgx.addChild(cyl);
		this.drawRangeLabel(tgx, this.uriOfPickableEndingOBJ, dataTypeColor);
	}

	private void addComplementOf(TransformGroup tgx) {
		// creating Range Appearance
		Appearance apparenza = this.createRangeItemAppearance(new Color3f(.9f,
				.9f, .9f));// SPHERE
		Appearance apparenza2 = this.createRangeItemAppearance(new Color3f(.1f,
				.1f, .1f));// TORUS

		float radius = VisualizationOptions.sR_sphereRadius * 1.35f;
		// creating the range 3D Icon
		Sphere sphere = new Sphere(radius, Sphere.ENABLE_APPEARANCE_MODIFY
				| Sphere.GENERATE_NORMALS, 50, apparenza);
		sphere.setCapability(Sphere.ALLOW_BOUNDS_READ);
		sphere.setPickable(true);
		sphere.setUserData(this.uriOfPickableEndingOBJ);
		tgx.addChild(sphere);

		Transform3D rot = new Transform3D();
		rot.rotX(Math.PI / 4);
		TransformGroup tg = new TransformGroup(rot);
		Torus loop = new Torus(radius / 5, radius, apparenza2);
		loop.setPickable(false);
		tg.addChild(loop);

		Transform3D rot2 = new Transform3D();
		rot2.rotX(-Math.PI / 4);
		TransformGroup tg2 = new TransformGroup(rot2);
		Torus loop2 = new Torus(radius / 5, radius, apparenza2);
		loop2.setPickable(false);
		tg2.addChild(loop2);

		tgx.addChild(tg);
		tgx.addChild(tg2);

		this.drawRangeLabel(tgx, "COMPLEMENT OF", this.black);

	}

	private void addUnionRange(TransformGroup tgx) {
		Appearance apparenza1 = this.createRangeItemAppearance(new Color3f(.0f,
				.2f, .6f));// BLU
		Appearance apparenza2 = this.createRangeItemAppearance(new Color3f(1f,
				.2f, .2f));// RED
		Appearance apparenza3 = this.createRangeItemAppearance(new Color3f(.0f,
				.8f, .4f));// GREEN
		Appearance apparenza4 = this.createRangeItemAppearance(new Color3f(.6f,
				.0f, .6f));// VIOLET

		// creating the range 3D Icon
		float k = VisualizationOptions.sR_sphereRadius * 2 / 3;// determines
																// the distance
																// between the 4
																// spheres
		BranchGroup bgIcon3d = this.createRelationBG();
		// Sphere 1
		Transform3D trans1 = new Transform3D();
		trans1.setTranslation(new Vector3f(0, k, 0));
		TransformGroup tg1 = new TransformGroup(trans1);
		Sphere sphere1 = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza1);
		sphere1.setPickable(true);
		sphere1.setUserData(this.uriOfPickableEndingOBJ);
		tg1.addChild(sphere1);
		// Sphere 2
		Transform3D trans2 = new Transform3D();
		trans2.setTranslation(new Vector3f(.86f * k, (-.5f * k), 0));
		TransformGroup tg2 = new TransformGroup(trans2);
		Sphere sphere2 = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza2);
		sphere2.setPickable(false);
		tg2.addChild(sphere2);
		// Sphere 3
		Transform3D trans3 = new Transform3D();
		trans3.setTranslation(new Vector3f((-.86f * k), (-.5f * k), 0));
		TransformGroup tg3 = new TransformGroup(trans3);
		Sphere sphere3 = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza3);
		sphere3.setPickable(false);
		tg3.addChild(sphere3);
		// Sphere 4
		Transform3D trans4 = new Transform3D();
		trans4.setTranslation(new Vector3f(0, 0, k));
		TransformGroup tg4 = new TransformGroup(trans4);
		Sphere sphere4 = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza4);
		sphere4.setPickable(false);
		tg4.addChild(sphere4);

		bgIcon3d.addChild(tg1);
		bgIcon3d.addChild(tg2);
		bgIcon3d.addChild(tg3);
		bgIcon3d.addChild(tg4);
		tgx.addChild(bgIcon3d);

		this.drawRangeLabel(tgx, "UNION", this.black);

	}

	private void addIntersectionRange(TransformGroup tgx) {
		Appearance apparenza1 = this.createRangeItemAppearance(new Color3f(0,
				1f, 0), .65f);// VIOLET
		Appearance apparenza2 = this.createRangeItemAppearance(new Color3f(1f,
				0, 0), .65f);// RED
		Appearance apparenza3 = this.createRangeItemAppearance(new Color3f(0,
				0, 1f), .65f);// GREEN
		Appearance apparenza4 = this.createRangeItemAppearance(new Color3f(.2f,
				.2f, .2f));// GRAY

		// creating the range 3D Icon
		float k = VisualizationOptions.sR_sphereRadius * 2 / 3;// determines
																// the distance
																// between the 4
																// spheres
		BranchGroup bgIcon3d = this.createRelationBG();
		// Sphere 1
		Transform3D trans1 = new Transform3D();
		trans1.setTranslation(new Vector3f(0, 0, k));
		TransformGroup tg1 = new TransformGroup(trans1);
		Sphere sphere1 = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza1);
		sphere1.setPickable(true);
		sphere1.setUserData(this.uriOfPickableEndingOBJ);
		tg1.addChild(sphere1);
		// Sphere 2
		Transform3D trans2 = new Transform3D();
		trans2.setTranslation(new Vector3f(.86f * k, 0, (-.5f * k)));
		TransformGroup tg2 = new TransformGroup(trans2);
		Sphere sphere2 = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza2);
		sphere2.setPickable(true);
		sphere2.setUserData(this.uriOfPickableEndingOBJ);
		tg2.addChild(sphere2);
		// Sphere 3
		Transform3D trans3 = new Transform3D();
		trans3.setTranslation(new Vector3f((-.86f * k), 0, (-.5f * k)));
		TransformGroup tg3 = new TransformGroup(trans3);
		Sphere sphere3 = new Sphere(VisualizationOptions.sR_sphereRadius,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza3);
		sphere3.setPickable(true);
		sphere3.setUserData(this.uriOfPickableEndingOBJ);
		tg3.addChild(sphere3);
		// Sphere 4
		Transform3D trans4 = new Transform3D();
		trans4.setTranslation(new Vector3f(0, 0, 0));
		TransformGroup tg4 = new TransformGroup(trans4);
		Sphere sphere4 = new Sphere(VisualizationOptions.sR_sphereRadius * .7f,
				Sphere.ENABLE_APPEARANCE_MODIFY | Sphere.GENERATE_NORMALS, 50,
				apparenza4);
		sphere4.setPickable(false);
		tg4.addChild(sphere4);

		bgIcon3d.addChild(tg1);
		bgIcon3d.addChild(tg2);
		bgIcon3d.addChild(tg3);
		bgIcon3d.addChild(tg4);
		tgx.addChild(bgIcon3d);

		this.drawRangeLabel(tgx, "INTERSECTION", this.black);
	}

	private void addRestrictionRange(TransformGroup tgx) {
		Appearance apparenza1 = this.createRangeItemAppearance(new Color3f(
				.27f, .27f, .27f));// FOG
		Appearance apparenza2 = this.createRangeItemAppearance(new Color3f(1f,
				0, 0));// RED
		Appearance apparenza3 = this.createRangeItemAppearance(new Color3f(
				.92f, .75f, .07f));// YELLOW
		Appearance apparenza4 = this.createRangeItemAppearance(new Color3f(
				.08f, .45f, .92f));// BLUE
		Appearance apparenza5 = this.createRangeItemAppearance(new Color3f(.2f,
				.8f, .2f));// GREEN
		Appearance apparenza6 = this.createRangeItemAppearance(new Color3f(.7f,
				.1f, .2f));// WINE

		// creating the range 3D Icon
		BranchGroup bgIcon3d = this.createRelationBG();
		float radius = VisualizationOptions.sR_sphereRadius;
		// Sphere 1
		TransformGroup tg1 = new TransformGroup();
		Sphere sphere1 = new Sphere(radius, Sphere.ENABLE_APPEARANCE_MODIFY
				| Sphere.GENERATE_NORMALS, 50, apparenza1);
		sphere1.setPickable(true);
		sphere1.setUserData(this.uriOfPickableEndingOBJ);
		tg1.addChild(sphere1);
		// Cone 1 UP
		Transform3D trans2 = this.computeOrientation(new Vector3f(0f,
				radius * .7f, 0f), 1.8f, false, false);
		TransformGroup tg2 = new TransformGroup(trans2);
		Cone cone1 = new Cone(VisualizationOptions.sR_coneRadius * .5f,
				VisualizationOptions.sR_coneHeight * .5f, Cone.GENERATE_NORMALS
						| Cone.GENERATE_TEXTURE_COORDS, apparenza2);
		cone1.setPickable(false);
		tg2.addChild(cone1);
		// Cone 2 CIRC
		Transform3D trans3 = this.computeOrientation(new Vector3f(radius * .7f,
				0, radius * .4f), 1.8f, false, false);
		TransformGroup tg3 = new TransformGroup(trans3);
		Cone cone2 = new Cone(VisualizationOptions.sR_coneRadius * .5f,
				VisualizationOptions.sR_coneHeight * .5f, Cone.GENERATE_NORMALS
						| Cone.GENERATE_TEXTURE_COORDS, apparenza3);
		cone2.setPickable(false);
		tg3.addChild(cone2);
		// Cone 3 CIRC
		Transform3D trans4 = this.computeOrientation(new Vector3f(
				-radius * .7f, 0, radius * .4f), 1.8f, false, false);
		TransformGroup tg4 = new TransformGroup(trans4);
		Cone cone3 = new Cone(VisualizationOptions.sR_coneRadius * .5f,
				VisualizationOptions.sR_coneHeight * .5f, Cone.GENERATE_NORMALS
						| Cone.GENERATE_TEXTURE_COORDS, apparenza4);
		cone3.setPickable(false);
		tg4.addChild(cone3);
		// Cone 4 DOWN
		Transform3D trans5 = new Transform3D();
		trans5.rotZ(Math.PI);
		trans5.setTranslation(new Vector3f(0, -radius * 1.3f, 0));
		TransformGroup tg5 = new TransformGroup(trans5);
		Cone cone4 = new Cone(VisualizationOptions.sR_coneRadius * .5f,
				VisualizationOptions.sR_coneHeight * .5f, Cone.GENERATE_NORMALS
						| Cone.GENERATE_TEXTURE_COORDS, apparenza5);
		cone4.setPickable(false);
		tg5.addChild(cone4);
		// Cone 5 CIRC
		Transform3D trans6 = this.computeOrientation(new Vector3f(0f, 0,
				-radius * .8f), 1.8f, false, false);
		TransformGroup tg6 = new TransformGroup(trans6);
		Cone cone5 = new Cone(VisualizationOptions.sR_coneRadius * .5f,
				VisualizationOptions.sR_coneHeight * .5f, Cone.GENERATE_NORMALS
						| Cone.GENERATE_TEXTURE_COORDS, apparenza6);
		cone5.setPickable(false);
		tg6.addChild(cone5);

		bgIcon3d.addChild(tg1);
		bgIcon3d.addChild(tg2);
		bgIcon3d.addChild(tg3);
		bgIcon3d.addChild(tg4);
		bgIcon3d.addChild(tg5);
		bgIcon3d.addChild(tg6);

		tgx.addChild(bgIcon3d);

		this.drawRangeLabel(tgx, "RESTRICTION", this.black);
	}

	private void addEnumerationRange(TransformGroup tgx) {
		Appearance apparenza1 = this.createRangeItemAppearance(new Color3f(.0f,
				.2f, .6f), .3f);// BLU
		Appearance apparenza2 = this.createRangeItemAppearance(new Color3f(1f,
				.1f, .1f));// RED
		Appearance apparenza3 = this.createRangeItemAppearance(new Color3f(.0f,
				.8f, .2f));// GREEN
		Appearance apparenza4 = this.createRangeItemAppearance(new Color3f(.6f,
				.0f, .6f));// VIOLET
		Appearance apparenza5 = this.createRangeItemAppearance(new Color3f(.5f,
				.25f, .0f));// BROWN
		Appearance apparenza6 = this.createRangeItemAppearance(new Color3f(.5f,
				.1f, .0f));// LIGHT GREEN

		// creating the range 3D Icon
		float side = VisualizationOptions.sR_cubeSide * .3f;// cube side
		float radius = VisualizationOptions.sR_sphereRadius * 1.3f;// sphere
																	// radius
		BranchGroup bgIcon3d = this.createRelationBG();
		// Sphere 1
		TransformGroup tg1 = new TransformGroup();
		Sphere sphere = new Sphere(radius, Sphere.ENABLE_APPEARANCE_MODIFY
				| Sphere.GENERATE_NORMALS, 50, apparenza1);
		sphere.setPickable(true);
		sphere.setUserData(this.uriOfPickableEndingOBJ);
		tg1.addChild(sphere);
		// Cube 1 up
		Transform3D trans2 = this.computeOrientation(new Vector3f(0f,
				radius * .7f, 0f), 1.3f, false, false);
		TransformGroup tg2 = new TransformGroup(trans2);
		Box cube1 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
				| Box.GENERATE_NORMALS, apparenza2);
		cube1.setPickable(false);
		tg2.addChild(cube1);
		// cube2
		Transform3D trans3 = this.computeOrientation(new Vector3f(radius * .7f,
				0, radius * .4f), 1.3f, false, false);
		TransformGroup tg3 = new TransformGroup(trans3);
		Box cube2 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
				| Box.GENERATE_NORMALS, apparenza3);
		cube2.setPickable(false);
		tg3.addChild(cube2);
		// cube 3
		Transform3D trans4 = this.computeOrientation(new Vector3f(
				-radius * .7f, 0, radius * .4f), 1.3f, false, false);
		TransformGroup tg4 = new TransformGroup(trans4);
		Box cube3 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
				| Box.GENERATE_NORMALS, apparenza4);
		cube3.setPickable(false);
		tg4.addChild(cube3);
		// cube 4 DOWN
		Transform3D trans5 = this.computeOrientation(new Vector3f(0,
				-radius * .7f, 0), 1.3f, false, false);
		TransformGroup tg5 = new TransformGroup(trans5);
		Box cube4 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
				| Box.GENERATE_NORMALS, apparenza5);
		cube4.setPickable(false);
		tg5.addChild(cube4);
		// cube 5 CIRC
		Transform3D trans6 = this.computeOrientation(new Vector3f(0f, 0,
				-radius * .7f), 1.3f, false, false);
		TransformGroup tg6 = new TransformGroup(trans6);
		Box cube5 = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
				| Box.GENERATE_NORMALS, apparenza6);
		cube5.setPickable(false);
		tg6.addChild(cube5);

		bgIcon3d.addChild(tg1);
		bgIcon3d.addChild(tg2);
		bgIcon3d.addChild(tg3);
		bgIcon3d.addChild(tg4);
		bgIcon3d.addChild(tg5);
		bgIcon3d.addChild(tg6);

		tgx.addChild(bgIcon3d);

		this.drawRangeLabel(tgx, "ONE OF", this.black);
	}

	/**
	 * @param transparency
	 *            The transparency to set.
	 */
	public void setTransparency(float transparency) {
		this.transparency = transparency;
	}

	/**
	 * @param fullURI
	 *            The fullURI to set.
	 */
	public void setFullURI(String fullURI) {
		this.uriOfPickableEndingOBJ = fullURI;
	}

	static public void setColorToRelation(Color c, String relationURI) {
		Relation3DObject.coloredRelations.put(relationURI, c);
	}

	static public void removeRelationColor(String relationURI) {
		Relation3DObject.coloredRelations.remove(relationURI);
	}

	static public Color getRelationColor(String relationURI) {
		return Relation3DObject.coloredRelations.get(relationURI);
	}

	public void setRelationType(GraphicalLinkTypeEnum relationType) {
		this.relationType = relationType;
	}
}
